"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import string

from typing import List, Iterable

from spil.util.log import info, debug
from spil.util.exception import SpilException
from spil.conf import sid_templates, leaf_keys  # type: ignore
from spil.sid.core.sid_resolver import sid_to_dicts
from spil import Sid
from spil.util.caching import lru_cache as cache


def extrapolate(sids: Iterable[str], as_sid: bool = False) -> Iterable[str | Sid]:
    """
    From an iterable containing leaf node paths, extrapolates all the subnode paths.

    This is useful when the data source quickly provides leaves only, and we want to construct child data.

    For example: the path "TEST/A/CHR/HERO/MOD/V001/W/avi"
    will generate: "TEST/A/CHR/HERO/MOD/V001/W", "TEST/A/CHR/HERO/MOD/V001", "TEST/A/CHR/HERO/MOD", etc.
    until "TEST".

    This is a pure string operation.

    Is as_sid is True, returns Sids instead of strings (the default). Sids are a bit slower.

    :param as_sid: if we want the returned values to be Sids, instead of strings.
    :param sids: generator
    :return:
    """

    debug(
        "Start Extrapolate {}".format(sids)
    )  # TODO: warning if a string or Sid is given, need iterable.

    generated = set()

    for sid in sids:

        sid = str(sid)

        generated.add(sid)
        if as_sid:
            yield Sid(sid)
        else:
            yield sid

        parts = sid.split("/")  # use sip
        for i, key in enumerate(reversed(parts[:-1]), 1):
            new_sid = "/".join(parts[: -1 * i])
            if new_sid in generated:
                break
            else:
                generated.add(new_sid)
                # print(new_sid)
                if as_sid:
                    yield Sid(new_sid)
                else:
                    yield new_sid


def expand(sid: str | Sid, do_extrapolate: bool = False) -> List[Sid]:
    """
    "Expand" means replacing a double wildcard "/**" by the possible amount of simple wildcards "/*", wherever possible.
    This allows for simpler searches on multiple types.

    Example:
        To find all movie files from a shot, we need to search for: hamlet/s/sq010/sh0020/*/*/*/mov
        hamlet/s/sq010/sh0020/**/mov is a simpler form.

    By default, expand only returns leaf types (do_extrapolate=False)
    hamlet/s/sq010/sh0020/** will be expanded to files, eg hamlet/s/sq010/sh0020/*/*/*/* and hamlet/s/sq010/sh0020/*/*/*/*/*

    If do_extrapolate is True, expand will return all possible intermediate types.
    Eg. hamlet/s/sq010/sh0020/** will also be expanded to hamlet/s/sq010/sh0020 (shot), and hamlet/s/sq010/sh0020/* (task), etc.
    This is not the default behaviour.

    This function expects a string (or sid) and converts to a list of typed Sids.

    Note: if there is nothing to expand (no "/**" in the string),
    the string is typed as-is (calling "simple_typing(sid)"), to return a list of typed Sids.

    Expand can only receive "typeable" sids:
    Aliases and "or" syntax are not allowed (they are handled before).

    Expand works like this:

    - First we get the root of the given Sid, before the "/**".
    We get the roots basetype ("asset", or "shot", etc) and deduce the "leaf" key.
    The leaf key is the last key we look up.

    - Then we look up all existing types (sid_templates).
    For each type, we get the amount of keys, and replace "/**" by "/*" x the amount of keys.
    That gives us a test sid, that we use to instantiate valid sids, using all possible templates again.

    - we continue with all the types we have not already tested.

    :param sid: string (or sid)
    :param do_extrapolate: boolean
    :return: List of typed Sids
    """

    debug(f"Expanding {sid}")
    sid = str(sid)

    if not sid.count("/**"):  # nothing to expand
        r = simple_typing(sid)
        debug("Nothing to expand for {}. Just casting to Sid set {}".format(sid, r))
        return r

    if sid.count("/**") > 1:
        raise SpilException("Can only expand once in a Sid.")

    if sid.count("?"):  # sid contains Query ending. We put it aside, and later append it back
        sid, query = sid.split("?", 1)
    else:
        query = ""

    root = sid.split("/**")[0]
    basetype = Sid(root).basetype
    debug("Basetype: {}".format(basetype))
    if not basetype:
        raise SpilException(
            f'The Search Sids "{sid}" root "{root}" cannot be typed, so it cannot be expanded. This is probably a configuration error.'
        )

    leaf_key = leaf_keys.get(basetype)
    # leaf_key = 'version'  # TODO: find an option to edit this depending on the Finder.
    if not do_extrapolate and not leaf_key:
        raise SpilException(
            f'Leaf key not defined for basetype: "{basetype}". Please check config.'
        )

    tested = []
    found = []
    result = []
    for key, template in sid_templates.items():
        debug("Checking " + key)
        if key in found:
            debug(".. Already checked {}, continue".format(key))
            continue
        debug(f'.. Checking key "{key}"')
        keys = list(string.Formatter().parse(template))
        if do_extrapolate or keys[-1][1] == leaf_key:
            count = len(keys) - 1
            current = sid.count("/")
            needed = count - current + 1
            test = sid.replace("/**", "/*" * needed)
            if test in tested:
                debug("... Already tested, continue")
                continue
            else:
                tested.append(test)
            debug("... Filled {}x* --> {}".format(needed, test))
            matching = sid_to_dicts(test)
            debug("... Got {}".format(matching))
            for __type, data in matching.items():
                debug(".... found :" + __type)
                found.append(__type)
                if data and (do_extrapolate or (list(data)[-1] == leaf_key)):
                    if query:
                        new_sid = Sid("{}:{}?{}".format(__type, test, query))
                    else:
                        new_sid = Sid(__type + ":" + test)
                    debug(".... appending: {}".format(new_sid.uri))
                    result.append(new_sid)
        else:
            debug('.. Type "{}" is not a leaf, and do_extrapolate is False, skipped.'.format(key))
            continue

    return sorted(list(set(result)))


@cache
def simple_typing(sid: str | Sid) -> List[Sid]:
    """
    Takes a given sid string (or Sid), typically containing search symbols, and returns a list of typed Sids.
    If the sid cannot be typed, returns a list with Sid(sid) inside.

    Args:
        sid:

    Returns:

    """
    _sid = str(sid)

    if _sid.count("?"):  # sid contains Query ending. We put it aside, and later append it back
        _sid, query = _sid.split("?", 1)
    else:
        query = ""

    root = _sid.split("/*")[0]
    basetype = Sid(root).basetype
    if not basetype:  # not typed, returning as is
        return [Sid(sid)]
        # raise SpilException('The Search Sids "{}" root "{}" cannot be typed, so it cannot be expanded. This is probably a configuration error.'.format(sid, root))

    result = []
    matching = sid_to_dicts(_sid)
    debug("Got {}".format(matching))
    for __type, data in matching.items():
        debug("found :" + __type)
        if query:
            new_sid = Sid("{}:{}?{}".format(__type, _sid, query))
        else:
            new_sid = Sid(__type + ":" + _sid)
        debug("appending: {}".format(new_sid.uri))
        result.append(new_sid)

    return list(set(result)) or [Sid(sid)]


if __name__ == "__main__":
    from pprint import pprint
    from spil import setLevel
    from spil.util.log import DEBUG

    setLevel(DEBUG)

    sid = "hamlet/**"
    r = expand(sid)
    pprint(r)



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional

import importlib
import inspect

from spil import conf
from spil.util.log import debug
from spil.conf.util import pattern_replacing
from spil.util.caching import lru_cache as cache


@cache
def get_path_config(name: Optional[str] = None) -> PathConfig:

    if not name:  # either name, or configured default, or first config_name entry
        name = conf.default_path_config or list(conf.path_configs.keys())[0]  # type: ignore

    config_module_name = conf.path_configs.get(name)  # type: ignore

    config = PathConfig(name, config_module_name)  # type: ignore

    debug(f'Getting: {config}')

    return config


class PathConfig:

    def __init__(self, name: Optional[str] = 'fs_conf', config_module_name: Optional[str] = None):

        self.name = name
        config_module_name = config_module_name or name

        try:
            self.module = importlib.import_module(config_module_name)  # type: ignore
        except ModuleNotFoundError as e:
            problem = conf.sid_conf_import_error_message.format(module=config_module_name)
            print(problem)
            raise Exception(problem)

        for name, value in inspect.getmembers(self.module):
            if name.startswith('__'):
                continue

            setattr(self, name, value)

        pattern_replacing(self.path_templates, self.key_patterns)  # type: ignore

    def __str__(self):
        return f"PathConfig: {self.name} / {self.module}"


if __name__ == '__main__':

    from pprint import pprint
    from spil.util.log import setLevel, INFO, info, DEBUG

    info('Tests start')

    setLevel(INFO)

    conf.set('default_path_config', None)
    get_path_config.cache_clear()

    pc = get_path_config()
    pprint(pc.path_templates.get('project_root'))

    conf.set('default_path_config', 'server')
    get_path_config.cache_clear()

    pc = get_path_config()
    pprint(pc.path_templates.get('project_root'))

    pc = get_path_config('local')
    pprint(pc.path_templates.get('project_root'))

    pc = get_path_config('server')
    pprint(pc.path_templates.get('project_root'))


# type: ignore
"""
Example File System resolver config_name.
Maps file paths to Sid components.
"""
from spil_sid_conf import key_patterns

# for test and demo purposes
from pathlib import Path
project_root_path = Path(__file__).parent / "data" / "testing" / "SPIL_PROJECTS" / "LOCAL" / "PROJECTS"
# Replace "project_root_path" by your own folder root, eg
# r'/home/mh/Desktop/SPIL_PROJECTS/LOCAL/PROJECTS'

path_templates = {

    'project_root':            project_root_path.as_posix(),

    # type asset
    'asset__file':             '{@project_root}/{project}/PROD/{type:ASSETS}/{assettype}/{asset}/{task}/{version}/{assettype}_{asset}_{task}_{state}_{version}.{ext:scenes}',
    'asset__movie_file':       '{@project_root}/{project}/PROD/{type:ASSETS}/{assettype}/{asset}/{task}/{version}/OUTPUT/{assettype}_{asset}_{task}_{state}_{version}.{ext:movies}',
    'asset__cache_file':       '{@project_root}/{project}/PROD/{type:ASSETS}/{assettype}/{asset}/{task}/{version}/OUTPUT/{assettype}_{asset}_{task}_{state}_{version}.{ext:caches}',

    'asset__version':          '{@project_root}/{project}/PROD/{type:ASSETS}/{assettype}/{asset}/{task}/{version}',
    'asset__task':             '{@project_root}/{project}/PROD/{type:ASSETS}/{assettype}/{asset}/{task}',
    'asset__asset':            '{@project_root}/{project}/PROD/{type:ASSETS}/{assettype}/{asset}',
    'asset__assettype':        '{@project_root}/{project}/PROD/{type:ASSETS}/{assettype}',

    'asset':                   '{@project_root}/{project}/PROD/{type:ASSETS}',

    # type shot
    'shot__file':              '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}/{sequence}_{shot}/{task}/{version}/{sequence}_{shot}_{task}_{state}_{version}.{ext:scenes}',
    'shot__movie_file':        '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}/{sequence}_{shot}/{task}/{version}/EXPORT/{sequence}_{shot}_{task}_{state}_{version}.{ext:movies}',
    'shot__cache_node_file':   '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}/{sequence}_{shot}/{task}/{version}/EXPORT/{sequence}_{shot}_{task}_{node}_{state}_{version}.{ext:caches}',
    'shot__cache_file':        '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}/{sequence}_{shot}/{task}/{version}/EXPORT/{sequence}_{shot}_{state}_{version}.{ext:caches}',

    'shot__version':           '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}/{sequence}_{shot}/{task}/{version}',
    'shot__task':              '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}/{sequence}_{shot}/{task}',
    'shot__shot':              '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}/{sequence}_{shot}',
    'shot__sequence':          '{@project_root}/{project}/PROD/{type:SHOTS}/{sequence}',
    'shot':                    '{@project_root}/{project}/PROD/{type:SHOTS}',

    # type project
    'project':                '{@project_root}/{project}',
}

path_templates_reference = 'project_root'

# Specific mappings not needed here
path_defaults = {
    'state': 'WORK',
}
sidkeys_to_extrakeys = {}
extrakeys_to_sidkeys = {}

path_mapping = {  # the resolver might get the keys by values. if the value exists multiple times, the first one is returned.

    'project': {
        'HAMLET': 'hamlet',
    },
    'type': {
        'ASSETS': 'a',
        'SHOTS': 's',
    },
    'state': {
        'WORK': 'w',
        'PUBLISH': 'p',
    },

    # Specific path mapping by type not needed here
}

# not needed here ('/3d/scenes/': '/*/scenes/')
search_path_mapping = {}

project_path_names = list(path_mapping.get('project').keys())

# we override elements already set in sid_conf
key_patterns = key_patterns.copy()

key_patterns['__'].update({
        '{state}':   r'{state:(WORK|PUBLISH|\*|\>)}',
        '{state:p}': r'{state:(PUBLISH|\*|\>)}',
        '{state:w}': r'{state:(WORK|\*|\>)}',
    })

key_patterns['t'].update({
        '{project}':        r'{project:(' + '|'.join(project_path_names) + r'|\*\>)}',
        '{type:SEQUENCES}': r'{type:(SEQUENCES|\*\>)}',
        '{type:ASSETS}':    r'{type:(ASSETS|\*\>)}',
    })

if __name__ == '__main__':

    from pprint import pprint

    pprint(globals())

    for key in path_templates:
        print(path_templates.get(key))



"""
This test runs tests on "data" Sid objects.
Data Sid objects contain attribute calls like
"sid.children()"
which are delegated to Finders and Getters (FindInAll and GetFromAll by default)

Uses the example sids.
"""
import random

# import spil  # config path bootstrap
from hamlet_scripts.example_sids import sids  # type: ignore
from spil.tests.utils.sid_data_tester import check_data_sids


def test_data_sids():

    random.seed(26)
    random.shuffle(sids)
    tests = sids[:20]
    check_data_sids(tests, reraise=True)


if __name__ == "__main__":

    from spil.util.log import setLevel, INFO
    setLevel(INFO)
    test_data_sids()



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, List, Dict, Optional, overload
from typing_extensions import Literal

from spil import Sid
from spil.sid.read.finder import Finder

from spil.util.log import debug, warning
from spil.conf import get_finder_for  # type: ignore  # dynamic config
from spil.util.caching import lru_cache as cache
from spil.sid.read.tools import unfold_search


@cache
def get_finder(sid: Sid | str, config: Optional[str] = None) -> Finder | None:
    """
    Calls spil.conf.get_finder_for() which is implemented in the spil_data_conf.

    Retrieves, for a given Search Sid and configuration name, the appropriate Finder and returns it.

    This typically returns a Finder depending on the Type of the Sid.
    For example: projects come from a FindInConstant, Assets and Shots may come from FindInPaths or FindInShotgrid.
    See the example spil_data_conf.get_finder_for()

    Technical note: the result is cached.
    This means that the choice of the Finder is cached, not the resulting data itself.
    The Finder is called again each time a query method (find(), find_one(), exists()) is called.

    Args:
        sid: the Search Sid for which we want to get the appropriate Finder instance
        config: an optional configuration name, to be able to have mutliple configs co-existing.

    Returns:
        the Finder to use for this Search Sid

    """
    _sid = Sid(sid)
    if not str(_sid) == str(sid):
        warning(f'Sid could not be instanced, likely a configuration error. "{sid}" -> {_sid}')
    source = get_finder_for(_sid, config)
    if source:
        debug(f'Getting data source for "{sid}": -> {source}')
        return source
    else:
        warning(f'Data Source not found for Sid "{sid}" ({_sid.type})')
        return None


class FindInAll(Finder):  # noqa
    """
    This Finder will call other Finders, as configured, depending on the search sids type.

    The do_find() method is delegated to other finders, and not implemented.
    """

    def __init__(self, config: Optional[str] = None):  # noqa
        """
        "config" is an argument that will be passed through to the config, via get_finder_for(sid, config).
        "config" acts like a key, to allow multiple FindInAll configurations to co-exist.

        Args:
            config: name of a configuration
        """
        self.config = config

    @overload
    def find(self, search_sid: str, as_sid: Literal[True]) -> Iterator[Sid]:  # noqa
        ...

    @overload
    def find(self, search_sid: str, as_sid: Literal[False]) -> Iterator[str]:  # noqa
        ...

    @overload
    def find(self, search_sid: Sid, as_sid: Literal[True]) -> Iterator[Sid]:  # noqa
        ...

    @overload
    def find(self, search_sid: Sid, as_sid: Literal[False]) -> Iterator[str]:  # noqa
        ...

    @overload
    def find(self, search_sid: str) -> Iterator[Sid]:  # noqa
        ...

    @overload
    def find(self, search_sid: Sid) -> Iterator[Sid]:  # noqa
        ...

    def find(
        self, search_sid: str | Sid, as_sid: Optional[bool] = True
    ) -> Iterator[Sid] | Iterator[str]:
        """
        Search dispatcher.

        The search is "unfolded" (one possibly untyped search sid is "unfolded" into a list of typed search sids).
        For each typed sid, we get the appropriate Finder.

        We then group the searches by Finder, to call each Finder with a list of searches (instead of each search individually).

        Args:
            search_sid: Sid to search
            as_sid: result will be returned as a Sid object if True, as a string otherwise.

        Returns:
            Generator over the found Sids, as Sid or string instances.
        """
        # we start by unfolding
        search_sids: List[Sid] = unfold_search(search_sid)

        # Dictionary to map a Finder to a list of Sids it should find.
        finder_to_searches: Dict[Finder, List[Sid]] = {}

        for search_sid in search_sids:

            # Finder for the specific search sid
            finder = get_finder(
                search_sid, self.config
            )  # TODO: allow multiple finders for the same search (eg.

            if finder:
                searches = finder_to_searches.get(finder) or []
                searches.append(search_sid)
                finder_to_searches[finder] = searches
            else:
                warning(f"No Finder configured for {search_sid}. Skipped from search.")

        done = set()
        for finder, searches in finder_to_searches.items():

            debug(f'Searching "{finder}" --> "{searches}"')
            generator = finder.do_find(searches, as_sid=False)

            for i in generator:
                # FIXME: why is data so often repeated, "if in done" is expensive, optimize
                if i not in done:
                    done.add(i)
                    if as_sid:
                        yield Sid(i)
                    else:
                        yield i

            else:
                debug(f'Nothing found for "{search_sid.uri}"')

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- Config: "{self.config}"]'


if __name__ == "__main__":
    print(FindInAll())



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil import Sid
from spil.conf import basetyped_search_narrowing  # type: ignore


def execute(sids):
    """
    Once a search Sid is typed, we may need to narrow down the values,
    (filling the string with values)
    to make sure the right type is hit in the search.

    The transformation cannot yet be properly automated.
    For that reason it is set in a config_name, and handled here.

    TODO: for more precise results we might want to implement this on types also,
    not only basetypes.

    Examples:
        >>> type_narrow(Sid("asset__assettype:hamlet/*/*"))
        Sid('asset__assettype:hamlet/a/*')

        >>> type_narrow(Sid("shot__sequence:hamlet/*/*"))
        Sid('shot__sequence:hamlet/s/*')

    Args:
        sids:

    Returns:

    """
    result = []

    for sid in sids:
        result.append(type_narrow(sid))

    return result


def type_narrow(sid: Sid) -> Sid:
    """
    For a given basetype, looks up the configured query, and applies it.

    Args:
        sid:

    Returns:

    """

    query = basetyped_search_narrowing.get(sid.basetype, "")
    if query:
        sid = sid.get_with(query=query)
    return sid


if __name__ == "__main__":

    from spil.util.log import DEBUG, setLevel

    # import doctest
    # doctest.testmod()
    setLevel(DEBUG)
    # sid = Sid("asset__assettype:hamlet/a/char")
    sid = Sid("asset__file:hamlet/a/char/hamlet/model/v001/p/*")
    s = type_narrow(sid)
    print(s.uri)

    sid = Sid("asset__movie_file:hamlet/a/char/hamlet/model/v001/p/*?type=~a")
    s = type_narrow(sid)
    print(s.uri)

    sids = [
        Sid("asset__file:hamlet/a/char/hamlet/model/v001/p/*"),
        Sid("asset__movie_file:hamlet/a/char/hamlet/model/v001/p/*"),
        Sid("asset__cache_file:hamlet/a/char/hamlet/model/v001/p/*"),
    ]

    done = execute(sids)
    print(done)



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
"""

        
        
        
        
        THIS IS WORK IN PROGRESS.
        
        
        
        

"""
from __future__ import annotations

import time
from multiprocessing import Process

from spil.sid.core.utils import extrapolate
from spil.sid.read.tools import unfold_search
from spil.util.caching import lru_kw_cache as cache

from spil.sid.read.finder import Finder

from pathlib import Path  # noqa

from spil import logging, SpilException, Sid, conf, FindInList

log = logging.get_logger(name="sid_cache")
log.setLevel(logging.ERROR)

EOL = "\n"
lock_outdate = 60 * 5  # 5 minutes
max_warmup_wait = 60  # 1 minute


def now():
    """Returns the 'now' timestamp as an int"""
    return int(time.time())


# @lru_cache
@cache
def get_sidcache(data_search, data_source, cache_file=None, name=None):
    """
    Factory to build pooled SidCache objects.

    "Pooled" means:
        Each sid_cache_file is handled by its own, single, SidCache instance.
        A SidCache object handles only one sid_cache_file.
        For example:


    Note that the file can be accessed concurrently from different python instances, from different machines.
    SidCache still needs to ensure threadsafety.

    #TODO: replace lru cache with a custom cache, with only filename as key.
    """
    return FindInCache(data_search=data_search, data_source=data_source, cache_file=cache_file, name=name)


class FindInCache(Finder):
    """
    Simple Sid cache implementation.
    It holds only Sids, not data related to Sids.
    It implements Data with methods: find, find_one, exists, create.

    There are 3 cached levels:
    - the FindInList ListSearch object, the cache in memory
    = a cache file on disk, the cache on disk for concurrent reads and updates
    = the original data source, that is to be cached.

    The cache can have 3 states:
    - hot: the Sids are cached in memory, in form of a FindInList (List Source) object.
      The clients can be served from the cache.
    - warm: the Sids are in a cache file on disk, but not yet loaded in memory, or outdated in memory.
      The cache needs to be loaded in memory, then it serves the client.
    - cold: no cache file exists, the file needs to be created and filled using the original data source.
      The SidCache will fill the file in the background (another process), and serve the client with the original data source.

    The workflow:
    - The cache initially reads the original data source into a file (warming up)
      During warmup, the cache serves the client via the original data source
    - When the file exists, it is read into memory (heating up)
      Now the cache is hot and serves clients
    - regularly, the cache checks if the file was updated since last read.
      If so, it re-reads it in memory.
    - The file is updated from concurrently existing SidCache instances.
      Thanks to the updates, the cache can be warm for a potentially indefinite time

    This cache implementation is very simple, and tailored to its working environment.
    It works well as long as:
    - Writes to the cache are rather rare, and reads are frequent
    - The cache is explicitly updated whenever the original data source is updated
      (so the cache can stay warm, because warming is slow)
    - The files are kept small, meaning the data is stored in multiple caches
      (under 5K items per cache)
    - Note: simultaneous writing of the same Sid may produce its duplication in the cache.

    The cache usage is simple and straightforward.
    get a new object via the factory function get_sidcache, by passing
    - a file path, for the cache file
    - a read sid and a data source, that will be the original data source.

    Typically, setting up the different caches for one or more projects is handled in the data_conf.
    The defined caches are then mapped to serve as data sources.

    Example:
        Setting up a cache on file '/temp/hamlet_shots.txt'
        For 'hamlet/s/*/*/*' (sequences, shots and tasks for hamlet project)
        Originally loaded from FS (File Search).
    ```
    ... sidcache = get_sidcache(sid_cache_file='/temp/hamlet_shots.txt', data_search='hamlet/s/*/*', data_source=FS())
    ... sidcache.find('hamlet/s/*')
    ```
    TODO: optional multiple data sources
    TODO: ttl for warmup
    TODO: file conform (dedoublon, sort) in background process
    """

    def __init__(self, data_search: str | Sid,
                 data_source: Finder,
                 cache_file: os.PathLike[str] = None,
                 name: str = None,
                 do_validate: bool = True):
        """
        cache_file is a file containing Sid strings, one per line (separated by EOL).
        cache_file should be a Path or path string.

        If do_cache_file is True (default) the file is converted to a list (loaded in RAM),
        else it is opened as a generator (reread at each usage).
        In case of large cache files, it is likely faster to set do_cache_file to False.

        If do_validate is True (the default), the data_source query is done with as_sid=True, to be certain only valid Sids make it in the cache.
        If you trust your data source, you can set it to False, which may slightly speed up the query.

        TODO: better error handling, to rule out problems in the input data (input, conf or cache).
        TODO: better newline handling. Maybe automatic strip in FindInList.
        TODO: add checks, for example that data_search on data_source finds results, if the cache_path exists and is writable, etc...
        """
        # defining entry
        self.data_search = str(data_search)

        self.name = name or f"sid_cache_{data_search.replace(conf.sip, '_').replace('*', '')}_{hash(data_search)}"

        self.cache_path = Path(str(cache_file)) or Path(conf.sid_cache_folder) / self.name + '.sids.txt'

        self.cache_list = []
        self.ls = None
        self.last_check_time = 0
        self.last_read_time = 0
        self.ttl = 30  # Minimal interval in seconds, before we check if the cache file modification time has been changed.
        self.do_validate = do_validate
        self.data_source = data_source

        # starting
        self._source()
        log.info("Init of SidCache with file {}".format(self.cache_path))

    def find(self, search_sid, as_sid=False):
        return self._source().find(search_sid, as_sid=as_sid)

    def do_find(self, search_sids, as_sid=False):
        return self._source().do_find(search_sids, as_sid=as_sid)

    def find_one(self, search_sid, as_sid=False):
        return self._source().find_one(str(search_sid), as_sid=as_sid)

    def exists(self, search_sid):
        return self._source().exists(str(search_sid))

    def create(self, sid, do_extrapolate=False):
        """
        Appends the given sid to cache_list and cache_file.

        Note on threadsafety.
        Write operations are kept atomic: either a simple append, or a complete replacement of the file.

        """
        if not self.cache_path.exists():
            log.debug("File does not exist, cache is cold, needs blocking warm up.")
            self._warm_up(blocking=True)
        self._heat_up()
        # this is not thread safe. It could happen that the data gets inserted twice. But it cannot be lost.
        if sid not in self.cache_list:
            with open(self.cache_path, "a") as f:
                f.write(str(sid) + EOL)
            self.cache_list.append(str(sid))
            self.ls = FindInList(searchlist=self.cache_list)

    def _source_to_file(self):
        """
        Reads the original data source into the cache file.

        """
        log.info("Start _source_to_file on {}".format(self.cache_path))
        lock = self.cache_path.with_suffix(".lock")
        temp_path = self.cache_path.with_name(
            "temp_{}".format(time.time()) + self.cache_path.suffix
        )
        generated = set()
        search_sids = unfold_search(self.data_search, do_extrapolate=True)
        from pprint import pformat
        log.error(pformat(search_sids))
        log.error(self.data_source)
        with open(temp_path, "a") as f:
            for sid in extrapolate(self.data_source.do_fing(search_sids, as_sid=self.do_validate), as_sid=self.do_validate):
                log.error(sid)
                if self.do_validate:
                    if not sid:
                        continue
                    else:
                        sid = str(sid)
                if sid in generated:
                    continue
                generated.add(sid)
                f.write(sid + EOL)
        try:
            Path(temp_path).replace(self.cache_path)
        except PermissionError:  # FIXME: better strategy
            time.sleep(5)
            Path(temp_path).replace(self.cache_path)
        lock.unlink()
        log.info("Done _source_to_file on {}".format(self.cache_path))

    @staticmethod
    def __wait_for_warmup(lock):
        """
        Waits as long as the given lock file exists.
        In case of time out, raises an exception.

        This is used by a SidCache that needs to wait for the warmup to complete, typically to insert data.
        """
        for i in range(max_warmup_wait):
            if lock.exists():
                time.sleep(1)
            else:
                return
        lock.unlink()
        raise SpilException(
            "Exhausted timeout ({} seconds) while waiting for cache warmup lock, file {}".format(
                max_warmup_wait, lock
            )
        )

    def clear(self, reload=True, blocking=False):
        if self.cache_path.exists():
            self.cache_path.unlink()
        if reload:
            self._warm_up(blocking=blocking)

    def _warm_up(self, blocking=False):
        """
        Inits the cache from the data source, from scratch.

        The warmup (cache reading function _source_to_file) is launched in a new process, protected by a lock file.

        If blocking is True, we wait for completion before returning.
        If blocking is False, we return the original data source, to serve the cache clients during warmup.
        """
        # check if no warmup already in process
        lock = self.cache_path.with_suffix(".lock")
        if lock.exists():
            if now() > os.path.getmtime(str(lock)) + lock_outdate:
                log.debug("Warmup lock is outdated, will be removed for warmup relaunch.")
                lock.unlink()
            else:
                log.debug("Warmup process already ongoing")
                if blocking:
                    log.debug("Waiting for warmup to finish before returning")
                    self.__wait_for_warmup(lock)
                else:
                    log.debug("Returning original data source")
                    return self.data_source

        lock.touch()
        p = Process(target=self._source_to_file)
        p.start()
        if blocking:
            # we do not p.join because we want to raise Exception if wait was too long
            self.__wait_for_warmup(lock)
            log.info("Warm_up finished for {}".format(self))
        else:
            log.info("Warm_up launched for {}".format(self))

    def _heat_up(self):
        """
        The cache file is loaded into memory.
        This process is blocking.
        """

        if not self.cache_path.exists():
            log.debug("File does not exist, cache is cold, needs warm up.")
            return self._warm_up()

        log.debug("File exists, reading file into memory (heating up)")
        self.cache_list = [line.rstrip() for line in open(self.cache_path, "r")]
        self.ls = FindInList(searchlist=self.cache_list)
        self.last_read_time = now()
        return self.ls

    def _source(self):
        """
        Checks for the Caches temperature, and returns the adequate data source.
        Either the FindInList object, if the cache is hot, or the original data source, if it is cold.

        Depending on the state of the cache, it launches heat-up prior to serving, or warm-up in the background.
        """
        log.debug("_source")
        if now() > self.last_check_time + self.ttl:
            log.debug("File date check is due")
            try:
                file_time = os.path.getmtime(str(self.cache_path))
                self.last_check_time = now()
            except FileNotFoundError:
                return self._warm_up()
            if file_time > self.last_read_time:
                return self._heat_up()

        return self.ls or self._heat_up()

    def __str__(self):
        return '<SidCache: "{}"@{} ({})>'.format(
            self.data_search, self.cache_path.name, self.data_source
        )

    def __repr__(self):
        return str(self)


if __name__ == "__main__":

    from spil_tests import stop
    import os

    log.setLevel(logging.DEBUG)

    sc = get_sidcache("V:/TESTPREMIERE/sdc.sids.txt", "FTOT/*/*")

    sc2 = get_sidcache("V:/TESTPREMIERE/ffm.sids.txt", "FFM/A,S,R/*/*/*")

    sc3 = get_sidcache("V:/TESTPREMIERE/sdc.sids.txt", "FTOT/*/*")

    assert sc == sc3

    print(get_sidcache.cache_info())
    for i in sc.find("FTOT/S/*"):
        print(i)

    for i in sc2.find("FFM/S/SQ0001/*"):
        print(i)

    # sc2.create('FFM/S/SQ0001/SH0007')

    for i in sc2.find("FFM/S/SQ0001/SH0007"):
        print(i)

    stop()
    # get_sidcache.cache_clear()

    print(list(Data().find("CBM/*")))

    print(list(Data().find("CBM/*")))

    print(list(Data().find("ARM/*")))

    # print(list(Data().find('FFM/A,S/*/*/*')))
    for i in Data().find("FFM/A,S/*/*/*"):
        print(i)

    sc4 = get_sidcache("V:/TESTPREMIERE/sdc.sids.txt", "CBM/*")

    get_sidcache.cache_info()



# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.

'''Custom error classes.'''


class ParseError(Exception):
    '''Raise when a template is unable to parse a path.'''


class FormatError(Exception):
    '''Raise when a template is unable to format data into a path.'''


class NotFound(Exception):
    '''Raise when an item cannot be found.'''


class ResolveError(Exception):
    '''Raise when a template reference can not be resolved.'''



# -*- coding: utf-8 -*-
# type: ignore
"""

This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
import importlib
import inspect

from spil.conf.util import extrapolate_templates, pattern_replacing

# stubs that are replaced by imports
sid_templates = {}
to_extrapolate = []
key_types = {}
extrapolate_types = {}
extrapolation_leaf_subtype = None
basetyped_search_narrowing = {}
key_patterns = {}
extension_alias = {}
projects = []
leaf_keys = {}

try:
    module = importlib.import_module('spil_sid_conf')
except ModuleNotFoundError as e:

    try:
        import sys
        from spil.conf import default_sid_conf_path, sid_conf_using_demo_configuration_message
        sys.path.append(default_sid_conf_path)
        module = importlib.import_module('spil_sid_conf')
        print(sid_conf_using_demo_configuration_message)

    except Exception as e:
        from spil.conf import sid_conf_import_error_message
        problem = sid_conf_import_error_message.format(module='spil_sid_conf')
        print(problem)
        raise Exception(problem)

__all__ = []
for name, value in inspect.getmembers(module):
    if name.startswith('__'):
        continue

    globals()[name] = value
    __all__.append(name)

sid_templates = extrapolate_templates(sid_templates, to_extrapolate)
pattern_replacing(sid_templates, key_patterns)

if __name__ == '__main__':

    from pprint import pprint

    pprint(globals())



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from urllib.parse import urlencode
from urllib import parse as urlparse

from spil.sid.core import sid_resolver
from spil import conf
from spil.util.log import warning, debug
from spil.util.exception import SpilException

"""
Query style string handling helper functions.

TODO: Sid Query handling needs to be cleaned up, typed and documented.
"""


def to_dict(query_string):
    """
    Converts a query string into a dictionary.

    All ? can be used as &.
    Optionally leading or rtailing ? or & are ignored.

    Examples:
    >>> to_dict('keyA=valueA&keyB=valueB, XX')
    {'keyA': 'valueA', 'keyB': 'valueB, XX'}

    >>> to_dict('?keyA=valueA&keyB=valueB')
    {'keyA': 'valueA', 'keyB': 'valueB'}

    >>> to_dict('?keyA=valueA?keyB=valueB')
    {'keyA': 'valueA', 'keyB': 'valueB'}

    """
    # we allow a=b?x=y but change it to a proper query: a=b&x=y
    query_string = query_string.replace("?", "&")

    # cleaning start / end signs
    if query_string.startswith("&"):
        query_string = str(query_string[1:])

    if query_string.endswith("&"):
        query_string = str(query_string[:-1])

    query_pairs = dict(urlparse.parse_qsl(urlparse.urlsplit("?" + query_string).query))
    return query_pairs


def to_string(query_dict):
    """
    Converts a query dict into a string.
    Does not do urlencoding, just strips whitespaces.

    Examples:
    >>> to_string({'keyA': 'valueA', 'keyB': 'valueB, BB'})
    'keyA=valueA&keyB=valueB,BB'
    """

    def encode(x, *args, **kwargs):
        return x.replace(" ", "")

    return urlencode(query_dict, quote_via=encode)  # noqa


def update(data, query, option_prefix="~"):
    """
    Updates given dict with given query into a new dict.

    If the Query value starts with option_prefix, it is only updated, if the key exists, not added.

    Option_prefix are removed from values, unless explicitely set to None.
    Keys and values are supposed to be strings.

    Examples:

        >>> update({'keyA': 'valueA', 'keyB': 'valueB'}, 'keyB=~replaceB&keyC=~skip this&keyD=keep this')
        {'keyA': 'valueA', 'keyB': 'replaceB', 'keyD': 'keep this'}

        >>> update({'keyA': 'valueA'}, 'keyB=~valueB', option_prefix=None)
        {'keyA': 'valueA', 'keyB': '~valueB'}

        >>> update({'keyA': 'valueA'}, 'keyB=valueB')
        {'keyA': 'valueA', 'keyB': 'valueB'}

    Args:
        data: a key:value dictionary
        query: a query string
        option_prefix: A sign prepended to a value, to indicate it is optional. "~"  by default.

    Returns: updated disctionary

    """
    if not data:  # may be None
        data = {}

    data = data.copy()
    new_data = to_dict(query)

    for key, value in new_data.items():

        if option_prefix:
            if str(value).startswith(str(option_prefix)):
                value = str(value).replace(option_prefix, "")
                is_value_optional = True
            else:
                is_value_optional = False
        else:
            is_value_optional = False

        if key in data.keys() or (not is_value_optional):
            data[key] = value

    return data


def apply_query(string, query=None, type=None, fields=None):
    """
    "Query application" algorithm:
    - the Query key/values update the existing fields dict --> query_helper.update
    - the updated fields dictionary is sent to the resolver to be typed --> sid_resolver.dict_to_type(fields, all=True)
    - if a single type is resolved, it is now used, integration done.
    - else if no type is resolved, the Query is not applied (kept in the string, the fields and type are unchanged)
    - else if multiple types are resolved:
        * if the original type is amongs the new types, it is used, integration done.
        * else if the sid is a search, the first type is used, integration done (#FIXME: this behaviour should change)
        * else the Query is not applied (kept in the string, the previous fields and type are unchanged)

    This is a problem in case of a read Sid, which could be poly-typed.

    Examples.
    (Note that the examples depend on "spil_hamlet_conf" example configuration. Failure may be due to non matching config)

    Query updates the sequence:
    >>> apply_query('hamlet/s/sq010', query='sequence=sq030', type='shot__sequence', fields={'project':'hamlet','type':'s','sequence':'sq010'})
    ('hamlet/s/sq030', 'shot__sequence', {'project': 'hamlet', 'type': 's', 'sequence': 'sq030'})

    Query adds a shot:
    >>> apply_query('hamlet/s/sq010', query='shot=sh0010', type='shot__sequence', fields={'project':'hamlet','type':'s','sequence':'sq010'})
    ('hamlet/s/sq010/sh0010', 'shot__shot', {'project': 'hamlet', 'type': 's', 'sequence': 'sq010', 'shot': 'sh0010'})

    Query updates the sequence and adds a shot:
    >>> apply_query('hamlet/s/sq010', query='sequence=*&shot=sh0010', type='shot__sequence', fields={'project':'hamlet','type':'s','sequence':'sq010'})
    ('hamlet/s/*/sh0010', 'shot__shot', {'project': 'hamlet', 'type': 's', 'sequence': '*', 'shot': 'sh0010'})

    Query contains undigestable fields (wrong sequence format), and is not applied:
    >>> apply_query('hamlet/s/sq010', query='sequence=fuzz', type='shot__sequence', fields={'project':'hamlet','type':'s','sequence':'sq010'})
    ('hamlet/s/sq010?sequence=fuzz', 'shot__sequence', {'project': 'hamlet', 'type': 's', 'sequence': 'sq010'})

    Works with empty string, fields and type
    >>> apply_query('', query='project=hamlet', type=None, fields=None)
    ('hamlet', 'project', {'project': 'hamlet'})
    """
    if not type and fields:
        raise SpilException("Query can only be applied on typed fields (or empty fields).")

    if not query:
        return string, type, fields

    _type = type

    new_data = update(fields, query)
    new_types = sid_resolver.dict_to_type(new_data, all=True)

    if not new_types:
        warning(f'[Sid] After Query apply, Sid "{string}" has no type. Query will not be applied.')
        string = "{}?{}".format(string, query)
        return string, _type, fields

    # After Query apply, if there is one single new type, we use this new type
    if len(new_types) == 1:
        _type = new_types[0]

    # if the Sid now matches multiple types, we have to choose
    elif len(new_types) > 1:

        # if the given type matches one of the new types, we keep the given type
        if _type in new_types:
            pass

        # if the given type does not match any of the new types
        else:
            # if the Sid is a search, we use the first of the new types.
            # TODO: this is arbitrary and should be warned or better handled.
            if any(s in "{}?{}".format(string, query) for s in conf.search_symbols):
                debug(
                    f'[Sid] After Query apply, Sid "{string}" matches different types: {new_types}. '
                    f"But it is a Search, so Query will be applied."
                )
                # Query will be applied, using the first new_types (# FIXME: refuse the temptation to guess)
                # In this case we should check if all the types result in the same Sid string, and return it, untyped.
                _type = new_types[0]
            else:
                warning(
                    f'[Sid] After Query apply, Sid "{string}" matches different types: {new_types}. '
                    f"Query will not be applied."
                )
                string = "{}?{}".format(string, query)
                return string, _type, fields

    # fields updated by Query is OK
    new_string = sid_resolver.dict_to_sid(new_data, _type)
    if new_string:
        return new_string, _type, new_data
    else:
        raise SpilException(
            f"Sid: [{string}?{query}] Query was correctly applied, but unable to resolve back to Sid"
        )


if __name__ == "__main__":
    """
    Launches doctest (test in the doc).
    """
    from spil.util.log import info, setLevel, INFO

    setLevel(INFO)

    info("Tests start")

    import doctest

    # info(doctest)

    doctest.testmod()

    r = to_dict("&keyA=valueA&keyB=valueB&")
    # print(r)

    info("Tests done.")



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2021 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
import sys

import logging
import logzero
from logzero import logger

"""
A simple logger shortcut / wrapper.

Uses
https://logzero.readthedocs.io/

TODO: complete code clean.
"""

#__logFormat = '[%(asctime)s] %(levelname)8s| %(message)s (%(filename)s:%(funcName)s:%(lineno)d)'

__logFormat = '[%(asctime)s] %(levelname)-6s| [%(module)s.%(funcName)s] %(message)-80s (%(lineno)d)'
__logFormat_colored = '%(color)s[%(asctime)s] %(levelname)-6s| [%(module)s.%(funcName)s] %(message)-80s (%(lineno)d)%(end_color)s'
#logging.basicConfig(format=__logFormat, level=conf.loglevel) # ??

# The following is needed to override the Maya logging configuration, the "basicConfig" does not work
#logger = logging.getLogger()
"""
for handler in logger.handlers:
    handler.setFormatter(logging.Formatter(fmt=__logFormat))
"""

# we set a new handler

handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logzero.LogFormatter(fmt=__logFormat, color=True))
logger.handlers = []
logger.addHandler(handler)

logzero.formatter(logging.Formatter(fmt=__logFormat))

logger.setLevel(logging.INFO)  # FIXME: config_name (depending on deploy dir)


def get_logger(name, color=True):
    new_logger = logzero.setup_logger(name=name)
    if color:
        new_logger.handlers[0].setFormatter(logzero.LogFormatter(fmt=__logFormat_colored, color=True))
    else:
        new_logger.handlers = []  # if we keep the handler, all logs are red.
        new_logger.addHandler(handler)
    return new_logger

"""
Code shortcuts
"""
debug = logger.debug
info = logger.info
warn = logger.warning  # warn is deprecated
warning = logger.warning
error = logger.error
critical = logger.critical

setLevel = logger.setLevel
getLevel = logger.getEffectiveLevel

DEBUG = logging.DEBUG
INFO = logging.INFO
WARN = logging.WARN
ERROR = logging.ERROR


if __name__ == '__main__':

    # specific logger (recommended)
    log_one = get_logger('one')
    # log_one = get_logger('one', color=False)
    log_one.debug('debug one')
    log_one.info('info one')
    log_one.setLevel(ERROR)
    log_one.debug('debug one')
    log_one.error('error one')

    # default logger
    setLevel(INFO)
    debug('titi')

    info('toto')
    warn('toto')
    error('toto')
    critical('toto')

    setLevel(DEBUG)

    debug('titi')


"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.conf import extension_alias  # type: ignore # from sid conf
from spil.conf import sip, ors
from spil.sid.core import query_helper


def execute(sids):
    """
    Maps extension aliases to a list of "or operator" separated extensions.

    Examples (with comma "," being the or-operator) :
         img -> jpg,exr,dpx
         hou -> hip, hipnc
         maya -> ma, mb
    """
    result = []

    for sid in sids:
        result.append(extensions(sid))

    return result


def handle_extension(extension_string=""):
    """
    Updates the string representing the extension, by replacing the aliases by the actual extensions.

    Examples:
    >>> handle_extension('maya')
    'ma,mb'

    >>> handle_extension('maya, mov, mb')
    'ma,mb,mov'

    >>> handle_extension()
    ''
    """
    if not extension_string:
        return ""

    if extension_string.count(ors):
        extension_string = [x.strip() for x in extension_string.split(ors)]
    else:
        extension_string = [extension_string]

    result = []
    for ext in extension_string:
        ext = extension_alias.get(ext, [ext])
        result.extend(ext)

    return ",".join(sorted(list(set(result))))


def extensions(sid):
    """
    Replaces extension aliases in the given string.
    Does replacement on the main string, and on an optional query.

    Examples
    >>> extensions('bla/s/bla/bla/**/hou')
    'bla/s/bla/bla/**/hip,hipnc'

    >>> extensions('bla/s/bla/bla/**/exr?ext=maya, mov')
    'bla/s/bla/bla/**/exr?ext=ma,mb,mov'

    >>> extensions('bla/s/bla/bla/**/movie, ma?ext=maya, mov')
    'bla/s/bla/bla/**/avi,ma,mov,mp4?ext=ma,mb,mov'

    #FIXME: "ext" is hard coded.
    """

    sid = str(sid)

    if sid.count("?"):  # sid contains Query ending. We put it aside, and later append it back
        sid, query = sid.split("?", 1)
    else:
        query = ""

    # main sid
    parts = sid.split(sip)
    newsid = parts[:-1]  # SMELL: this might not be an extension at all. Still works.
    newsid.append(handle_extension(parts[-1]))

    # query part
    if query:
        query_dict = query_helper.to_dict(query)
        if query_dict.get("ext"):  # FIXME: "ext" is hard coded.
            query_dict["ext"] = handle_extension(query_dict.get("ext"))
        query = query_helper.to_string(query_dict)

    return sip.join(newsid) + ("?" + query if query else "")


if __name__ == "__main__":

    import doctest

    doctest.testmod()

    """
    from spil.util.log import setLevel, info
    from logging import FATAL
    info('Tests start')
    setLevel(FATAL)
    expandables = ['hamlet/s/*/hou',
                   'hamlet/s/*/p/*/exr,img, hou?test',
                   'hamlet',
                   'hamlet/a/*/img?test',
                   'hamlet/s/sq010/sh0010/anim/*/v001/p/vdb',
                   'hamlet/s/*/movie',
                   'hamlet/s/sq010/sh0010/**/exr?ext=maya, mov']

    for sid in expandables:
        print(sid + '-->' + str(extensions(sid)))
        print('*'*10)
    """



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2021 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.


"""


class SubStr(str):
    """
    A string that knows substraction ... 
    """
    
    def __sub__(self, other):
        """
        "Substracts" - meaning removes from the right outer part - other from self. 
        """
        try:
            if other:
                return SubStr( ''.join(self.rsplit(str(other),1)) )
            else:
                return self
        except Exception as ex:
            print('so you really believed string substraction could work ? ',)
            print(ex)
            return self
        
        
if __name__ == '__main__':

    # definitions
    a = 'anything'
    x = 'any other string'
    y = 'not the end of a'
    assert(SubStr(a + x) - x == a)      # a + x - x == a
    assert(SubStr(a) - a == '')         # a - a = ''
    if not a.endswith(y):
        assert(SubStr(a) - y == a)

    s = SubStr('hello you')
    print( s - 'me' )
    print( s - 'you' )
    print( s - '' )
    print( s - None )
    print( 's - s : ', s - s )
    print( 's-s-"you" : ', s - s - 'you' )



sip = '/'  # sid separator

projects = ['hamlet']

sid_templates = {

    # type asset
    'asset__file':            '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:scenes}',
    'asset__movie_file':      '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:movies}',
    'asset__cache_file':      '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:caches}',

    'asset__state':           '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}',  # extrapolated

    'asset':                  '{project}/{type:a}',

    # type shot
    'shot__file':             '{project}/{type:s}/{sequence}/{shot}/{task}/{version}/{state}/{ext:scenes}',
    'shot__movie_file':       '{project}/{type:s}/{sequence}/{shot}/{task}/{version}/{state}/{ext:movies}',
    'shot__cache_file':       '{project}/{type:s}/{sequence}/{shot}/{task}/{version}/{state}/{ext:caches}',
    'shot__cache_node_file':  '{project}/{type:s}/{sequence}/{shot}/{task}/{version}/{state}/{node}/{ext:caches}',
    'shot__cache_node':       '{project}/{type:s}/{sequence}/{shot}/{task}/{version}/{state}/{node}',

    'shot__state':            '{project}/{type:s}/{sequence}/{shot}/{task}/{version}/{state}',  # extrapolated

    'shot':                   '{project}/{type:s}',

    # type project
    'project':                '{project}',

}

to_extrapolate = ['asset__state', 'shot__state']

asset_tasks = ['art', 'model', 'surface', 'rig']
shot_tasks = ['board', 'layout', 'anim', 'fx', 'render', 'comp']
asset_types = ['char', 'location', 'prop', 'fx']

# alias as last value
extensions_scene = ['ma', 'mb', 'hip', 'blend', 'hou', 'psd', 'nk', 'maya']
extensions_cache = ['abc', 'json', 'fur', 'grm', 'vdb', 'cache']
extensions_movie = ['mp4', 'mov', 'avi', 'movie']

extension_alias = {
    'cache': extensions_cache[:-1],  # we remove the last value, which is the alias itself
    'hou': ['hip', 'hipnc'],
    'maya': ['ma', 'mb'],
    'movie': extensions_movie[:-1]
}

key_patterns = {

    '__': {
        '{state}':      r'{state:(w|p|\*|\>)}',         # "w" or "p" or *
        '{state:w}':    r'{state:(w|\*|\>)}',           # "w" or "p" or *
        '{state:p}':    r'{state:(p|\*|\>)}',           # "w" or "p" or *
        '{version}':    r'{version:(v\d\d\d|\*|\>)}',   # "V" followed by 3 digits, or *
        '{sequence}':   r'{sequence:(sq\d\d\d|\*|\>)}',   # "sq" followed by 3 digits, or *  # !!!: do not use r'{2}', error with lucidity
        '{shot}':       r'{shot:(sh\d\d\d\d|\*|\>)}',     # "sh" followed by 4 digits, or *  # !!!: do not use r'{3}', error with lucidity

        '{ext:scenes}': r'{ext:(' + '|'.join(extensions_scene) + r'|\*|\>)}',
        '{ext:caches}': r'{ext:(' + '|'.join(extensions_cache) + r'|\*|\>)}',
        '{ext:movies}': r'{ext:(' + '|'.join(extensions_movie) + r'|\*|\>)}',
    },

    'asset__': {
        '{task}': r'{task:(' + '|'.join(asset_tasks) + r'|\*|\>)}',
        '{assettype}': r'{assettype:(' + '|'.join(asset_types) + r'|\*|\>)}',
    },
    'shot__': {
        '{task}': r'{task:(' + '|'.join(shot_tasks) + r'|\*|\>)}',
    },
    't': {  # everything   containing a "t" (asset, shot, project...) # FIXME
        '{project}': r'{project:(' + '|'.join(projects) + r'|\*|\>)}',
        '{type:a}': r'{type:(a|\*|\>)}',
        '{type:s}': r'{type:(s|\*|\>)}',
    },
}

# These keytypes are used by the resolver to sort the keys
key_types = {
    'asset': ['project', 'type', 'assettype', 'asset', 'task', 'version', 'state', 'ext'],
    'shot': ['project', 'type', 'sequence', 'shot', 'task', 'version', 'state', 'node', 'ext'],
    'project': ['project'],
}

# "leaf": last key of a Sid, per basetype. Typically "ext".
leaf_keys = {'asset': 'ext',
             'shot': 'ext',
             'project': 'ext',  # used for expand
             None: 'ext'}  # if the Sid has None basetype

"""
Once a search Sid is typed, we may need to narrow down the values, to make sure the right type is hit in the search.
For example: 
"asset__assettype:hamlet/*/*" and "shot__sequence:hamlet/*/*" should be narrowed down to 
"asset__assettype:hamlet/a/*" and "shot__sequence:hamlet/s/*"
The patterns cannot yet be properly automated. The goal for this config_name is to explicitly configure this.
"""
basetyped_search_narrowing = {

    'asset': 'type=~a',
    'shot': 'type=~s',

}

typed_search_narrowing = {  # type: ignore
    # not implemented yet.
}


if __name__ == '__main__':

    from pprint import pprint

    pprint(globals())



# type: ignore
"""
This script reads the Sids that have been saved to the example sid file.
The Sids are loaded into the "sids" variable to be used in tests.

The script can also launch a "test_config_coverage".
This shows all the types covered by the test sids,
and which types are not.
"""
from hamlet_scripts.save_examples_to_file import sid_file

sids = []
with sid_file.open() as f:
    sids = f.read().splitlines()

if __name__ == "__main__":

    import sys
    from spil.tests import Timer

    from spil.util.log import DEBUG, get_logger

    log = get_logger("spil_tests")
    log.setLevel(DEBUG)

    from spil import Sid

    print("Start")
    # pprint(sids)

    print(len(sids))

    if False:  # set False to skip
        test_config_coverage(sids)

    y = input("print out detail ?")

    if y != "y":
        sys.exit()

    global_timer = Timer(name="global", logger=log.debug)
    global_timer.start()
    for i, s in enumerate(sids):
        sid = Sid(s)
        if not sid:
            print(f"---------------------------------> {sid}")
            sleep(1)
        if i % 20 == 0:
            print(
                f"{i} -- {repr(sid)}",
            )

    global_timer.stop()
    print("Done")

    print(len(sids))

    global_timer = Timer(name="global", logger=log.debug)
    global_timer.start()
    for i, s in enumerate(sids):
        sid = Sid(s)
        if not sid:
            print(f"---------------------------------> {sid}")
            sleep(1)
        if i % 20 == 0:
            print(
                f"{i} -- {repr(sid)}",
            )
    global_timer.stop()



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional, Mapping, Any

import shutil
import json

from spil import Sid, Writer
from spil.util.exception import SpilException
from spil.conf import default_path_config, create_file_using_template, create_file_using_touch, get_data_json_path  # type: ignore
from spil.util.log import debug, warning, error
from pathlib import Path


def _create_parent(path: Path) -> bool:
    """
    Creates parent folder(s) of given Path.

    Args:
        path: path for which parent should be created.

    Returns: True on success, False on failure.
    """

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    return path.parent.exists()


def _write_data(sid_path: Path, data: Mapping[str, Any]) -> bool:
    """
    Writes given data to json.
    To get the jsons file path, the suffix is replaced, as per config.

    If the json already exists, it is loaded, updated, and dumped.
    Else the json is created.

    Args:
        sid_path: path from a valid sid
        data: data dictionary

    Returns: True on success, False on failure
    """

    data_path = get_data_json_path(sid_path)

    try:
        # if there is already data, we load and update it
        if data_path.exists():
            with data_path.open() as f:
                previous_data = json.load(f) or {}
                previous_data.update(data)
                data = previous_data

        # dumping data, converting to string if not serializable
        data_path.write_text(json.dumps(data, indent=4, default=str))

        return data_path.exists()

    except json.JSONDecodeError as e:
        error(f"Could not decode json at {sid_path}. Error: {e}")
        raise
        # return False

    except OSError as e:
        error(f"Could not access the json file {sid_path}. Error: {e}")
        raise
        # return False

    except Exception as e:
        error(f"Unexpected exception saving data at {sid_path}. Error: {e}")
        raise
        # return False


class WriteToPaths(Writer):
    """
    Writer to Paths.
    "Writers" implement the CRUD's Create/Update/Delete functions.

    WriteToPaths edits Sids and Sid data at the file system.

    This class should be considered an example implementation.
    Although it can be considered useful, it is not by itself production ready.
    It is used to create mock/test files and folders for Sid testing purposes.

    As FindInPaths and the sid.path() it is possible to instantiate this Writer with a fs config name.
    The config name points to the path templates config file.

    Sids are created at the configured path.
    Sid File creation is either done by copy of a template file (if exists as configured),
    or by "touch", which creates an empty file.

    Data is created and updated in the form of a json file, at the same path, with a specific configured suffix.
    """

    def __init__(self, config: Optional[str] = None):
        self.config = config or default_path_config  # type: ignore

    def create(self, sid: Sid | str, data: Optional[dict] = None) -> bool:

        _sid = Sid(sid)
        if not _sid.path(self.config):
            raise SpilException(
                f"[WriteToPaths] Cannot Create: sid {sid} has no path (config: {self.config})."
            )

        path = _sid.path(self.config)
        if path.exists():
            raise SpilException(
                f"[WriteToPaths] Cannot Create: path already exists for {sid}. Path: {path} (config: {self.config})."
            )

        suffix = path.suffix
        if suffix:
            debug(f"Path is a file: {path}")
            template = create_file_using_template.get(suffix[1:])  # we remove the dot of the suffix
            if template:
                debug(f"Will be created by copying template: {template}")
                _create_parent(path)
                shutil.copy2(template, path)
            elif create_file_using_touch:
                debug(f"Will be created by touch. {path}")
                _create_parent(path)
                path.touch()
            else:
                debug(f"File will not be created: {path}")

        else:
            debug(f"Path is a directory {path}")
            path.mkdir(parents=True)

        if not path.exists():
            warning(f"Creation of {sid} has failed. Path was not created: {path}")
            return False

        # Now creating optional data
        if data:
            return _write_data(path, data)

        return path.exists()

    def update(self, sid: Sid | str, data: Mapping[str, Any]) -> bool:
        """
        Updates given data on given Sid.

        Args:
            sid: Sid string or instance
            data: data dictionary

        Returns: True on success, False on failure
        """
        _sid = Sid(sid)
        if not _sid.path(self.config):
            raise SpilException(
                f"[WriteToPaths] Cannot update: sid {sid} has no path (config: {self.config})."
            )

        path = _sid.path(self.config)
        if not path.exists():
            raise SpilException(
                f"[WriteToPaths] Cannot Update: Sid {sid} does not exist at path: {path} (config: {self.config})."
            )

        return _write_data(path, data)

    def set(
        self, sid: Sid | str, attribute: Optional[str] = None, value: Optional[Any] = None, **kwargs
    ) -> bool:
        """
        Sets given attribute with given value, and / or given kwargs, for given Sid.
        Updates the data if it already exists.

        Internally calls update().

        Args:
            sid: Sid string or instance
            attribute: attribute name, data key
            value: value for data
            **kwargs: key / value pairs.

        Returns: True on success, False on failure
        """
        if attribute:
            kwargs[attribute] = value
        return self.update(sid, data=kwargs)

    def delete(self, sid: Sid | str) -> bool:
        raise NotImplementedError("Deletion is currently not implemented.")

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- Config: "{self.config}"]'


if __name__ == "__main__":

    sid = "hamlet/a"
    data = {"author": "John"}
    print(WriteToPaths())
    writer = WriteToPaths()
    done = writer.update(sid, data)
    print(Sid(sid).path())
    print(f"Update: {done}")

    done = writer.set(sid, bill=25, random_attribute="random value")
    print(f"Set: {done}")

    done = writer.set(sid, attribute="bill", value=299)
    print(f"Set: {done}")



# -*- coding: utf-8 -*-
"""

This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
from typing import Dict
import importlib
import inspect

# stubs that are replaced by imports
get_finder_for = None
get_getter_for = None
sid_cache_folder = ''
sid_cache_path = ''
path_configs: Dict[str, str] = {}
default_path_config = ''

try:
    module = importlib.import_module('spil_data_conf')
except ModuleNotFoundError as e:
    from spil.conf import sid_conf_import_error_message
    problem = sid_conf_import_error_message.format(module='spil_data_conf')
    print(problem)
    raise Exception(problem)

__all__ = []
for name, value in inspect.getmembers(module):
    if name.startswith('__'):
        continue

    globals()[name] = value
    __all__.append(name)


if __name__ == '__main__':

    from pprint import pprint

    pprint(globals())



"""

This is the config file for data access.

It is ingested by spil.conf.data_conf_load, which reads it into spil.conf.


*This is beta / work in progress*

The data config specification is subject to change.
(without affecting spil's API).

"""
from __future__ import annotations
from pathlib import Path

#########################################################
# Path config
path_configs = {'local': 'spil_fs_conf',
                'server': 'spil_fs_server_conf'
                }
default_path_config = 'local'
#########################################################


#########################################################
# Config for FindInAll
def get_finder_for(search_sid, config=None):  # get finder by Sid and optional config
    """
    Configuration used by FindInAll, to define which Finder is used for a given Search Sid.

    For a given Search Sid, and optional config,
    looks up the Search Sid's type and the matching Finder, as defined in an ad-hoc dict.

    The "config" allows to create multiple FindInAll instances / behaviours.
    (similar to the FindInPaths where we Find in different file systems)

    Args:
        search_sid: sid that is searched by FindInAll
        config: optional config to be able to define multiple FindInAll instances.

    Returns:
        A Finder instance for this search.
    """
    # type: ignore
    from spil_sid_conf import projects, asset_types  # type: ignore
    from spil import FindInConstants, FindInPaths, Finder

    finder_paths = FindInPaths()
    finder_projects = FindInConstants("project", projects)
    finder_types = FindInConstants("type", ["a", "s"], parent_source=finder_projects)
    finder_assettypes = FindInConstants('assettype', asset_types, parent_source=finder_types)
    finder_asset_states = FindInConstants('state', ["w", "p"], parent_source=finder_paths)

    finders_by_type = {
        'project': finder_projects,
        'asset': finder_types,
        'shot': finder_types,
        'asset__assettype': finder_assettypes,
        'asset__state': finder_asset_states,
        'shot__state': finder_asset_states,
        'default': finder_paths
    }

    finder: Finder = finders_by_type.get(search_sid.type, {}) or finders_by_type.get('default', {})
    if finder:
        return finder
    else:
        return None


#########################################################
# Config for GetFromAll
def get_getter_for(sid, attribute=None, config=None):
    """
    Configuration used by GetFromAll, to define which Getter is used for a given Sid or Search Sid.

    For a given Sid, and optional attribute and config,
    looks up the Sid's type and the matching Getter, as defined in an ad-hoc dict.

    The "config" allows to create multiple GetFromAll instances / behaviours.
    (similar to the FindInPaths where we Get from different file systems)

    The function may return None, in which case the getter will not return anything.

    Args:
        sid: sid that is queried by GetFromAll
        attribute: optional attribute name
        config: optional config to be able to define multiple GetFromAll instances.

    Returns:
        A Getter instance for this search, or None is none is defined.
    """
    # from spil_action.libs.files import get_comment, get_size, get_time
    from spil import Getter, GetFromPaths
    # from spil_plugins.sg.get_sg import GetFromSG
    from hamlet_plugins.next_get import NextGetter

    attribute_getters = {
        "next.version": NextGetter()
        #'comment': get_comment,
        #'size': get_size,
        #'time': get_time,
    }

    getter: Getter | None = attribute_getters.get(attribute)
    if getter:
        return getter

    getters_by_type = {
        'project': None,
        'asset': None,
        'shot': None,
        'asset__assettype': None,
        'asset__state': None,
        'shot__state': None,
        # 'asset__asset': GetFromSG(),
        # 'shot__shot': GetFromSG(),
        # 'shot__sequence': GetFromSG(),
        # 'shot__task': GetFromSG(),
        # 'asset__task': GetFromSG(),
        'default': GetFromPaths()
    }

    if sid.type in getters_by_type:
        # getter can be explicitly None
        # in this case, no data can be retrieved.
        return getters_by_type.get(sid.type)
    else:
        return getters_by_type.get('default')


def get_writer_for(sid):
    """
    Configuration used by WriteToAll, to define which Writer is used for a given Sid.

    *This is work in progress*
    """
    raise NotImplementedError("get_writer_for is not implemented")


#########################################################
# Config for WriteToPaths & GetFromPaths
#
path_data_suffix = '.data.json'
# WriteToPath: create. If a template exists for a suffix (extension), we copy it.
create_file_using_template = {  # type: ignore
    # 'ma': '.../empty.ma',
    # 'mb': '.../empty.mb'
}
# WriteToPath: create. If no template exists, we create an empty file with path.touch(), if create_using_touch is True
create_file_using_touch = True
# If nothing of these is set, we do not create a file.


# Used by WriteToPaths and GetFromPaths
def get_data_json_path(sid_path: Path) -> Path:
    """
    For a given Sid path, returns the path of a hidden data "sidecar" json file.
    Adds a dot before the name and changes the extension to ".data.json"

    Args:
        sid_path: path of a Sid

    Returns:
        path of a hidden data sidecar json file.

    """

    # TODO: add file rotation
    data_path = sid_path.with_name('.' + sid_path.name).with_suffix(path_data_suffix)
    return data_path

# End of Config for WriteToPaths & GetFromPaths
#########################################################




# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2022 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""

"""
Global configuration test.

Shows what may be duplicate, and what is missing from either config_name (sid, fs)

"""
import spil  # default config bootstrap

from spil.conf import sid_templates  # type: ignore
from spil.sid.pathops.pathconfig import get_path_config
path_config = get_path_config()
path_templates = path_config.path_templates

from spil.util.log import DEBUG, ERROR, get_logger
log = get_logger('tests')
log.setLevel(DEBUG)


def test_show_config():

    log.debug('Starting')
    log.debug('Path Templates: ')
    for k, v in path_templates.items():
        log.info('{} -> {}'.format(k, v))


def get_duplicates(items):
    unique = set()
    duplicate = set()
    for item in items:
        if item not in duplicate:
            unique.add(item)
        else:
            duplicate.add(item)
    return duplicate


def test_fs_duplicates():

    log.debug('- Testing duplicates in path_templates (fs_conf)...')

    duplicate_keys = get_duplicates(path_templates.keys())
    if duplicate_keys:
        log.warning('\tFAILED: Duplicate keys in path_templates (fs_conf): {}'.format(duplicate_keys))
    else:
        log.info('\tOK: No duplicate in path_templates keys (fs_conf).')

    duplicate_values = get_duplicates(path_templates.values())
    if duplicate_values:
        log.warning('\tFAILED: Duplicate values in path_templates (fs_conf): {}'.format(duplicate_values))
    else:
        log.info('\tOK: No duplicate in path_templates values (fs_conf).')

    log.debug('done')


def test_missing():

    log.debug('- Testing missing in sid_conf vs fs_conf...')

    sid_keys = set(sid_templates.keys())
    fs_keys = set(path_templates.keys())

    to_ignore = {'project_root'}

    missing_in_fs_conf = sid_keys - fs_keys - to_ignore
    if missing_in_fs_conf:
        log.warning('\tFAILED: Missing sid_conf keys in FS keys: {}'.format(missing_in_fs_conf))
    else:
        log.info('\tOK: all sid_conf keys are in FS keys.')

    missing_in_sid_conf = fs_keys - sid_keys - to_ignore
    if missing_in_sid_conf:
        log.warning('\tFAILED: Missing FS keys in sid_conf keys: {}'.format(missing_in_sid_conf))
    else:
        log.info('\tOK: All FS keys are in sid_conf keys.')

    log.debug('done')


if __name__ == '__main__':

    log.debug('Starting')
    log.debug('Path Templates: ')
    for k, v in path_templates.items():
        log.info('{} -> {}'.format(k, v))

    log.debug('')
    log.debug('Tests:')
    test_fs_duplicates()
    test_missing()

    log.debug('Done')



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, List

from spil import Sid, Finder
from spil.sid.read.finders.find_glob import FindByGlob
from spil.util.exception import SpilException
from spil.util.log import debug


class FindInConstants(FindByGlob):
    """
    Finds Sid from a list, for a specific type.

    Whereas FindInList looks inside a list of Sids, FindInConstants looks inside a list for one specific keytype.

    Implements star_search for a specific key, by getting the values from a given, constant, list.
    If there is a search "above" the given key, the search is delegated to the given "parent_source".

    This is used to create a Data Source for constant values.

    Examples:

        We instantiate a parent Finder

            >>> from spil import FindInPaths
            >>> fp = FindInPaths()

        Finding a constant list of asset types

            >>> finder = FindInConstants('assettype', ['char', 'location', 'prop', 'fx'], parent_source=fp)
            >>> list( finder.find('hamlet/a/*') )
            [Sid('asset__assettype:hamlet/a/char'), Sid('asset__assettype:hamlet/a/location'), Sid('asset__assettype:hamlet/a/prop'), Sid('asset__assettype:hamlet/a/fx')]

        Search "above" in the hierarchy is delegated to the parent

            >>> list( finder.find('hamlet/*/prop') )
            [Sid('asset__assettype:hamlet/a/prop')]

        Other examples, defining a constant list of "states" and searching for states

            >>> finder_states = FindInConstants('state', ["w", "p"], parent_source=fp)
            >>> list( finder_states.find('hamlet/a/location/ramparts/rig/v001/*') )
            [Sid('asset__state:hamlet/a/location/ramparts/rig/v001/w'), Sid('asset__state:hamlet/a/location/ramparts/rig/v001/p')]

        Finding a state with a parent search

            >>> list( finder_states.find('hamlet/a/location/ramparts/rig/*/p') )
            [Sid('asset__state:hamlet/a/location/ramparts/rig/v001/p'), Sid('asset__state:hamlet/a/location/ramparts/rig/v002/p')]

        Finding a state with a multi-parent search

            >>> sorted(list( finder_states.find('hamlet/a/location/*/rig/*/p') ))
            [Sid('asset__state:hamlet/a/location/elsinore/rig/v001/p'), Sid('asset__state:hamlet/a/location/elsinore/rig/v002/p'), Sid('asset__state:hamlet/a/location/garden/rig/v001/p'), Sid('asset__state:hamlet/a/location/garden/rig/v002/p'), Sid('asset__state:hamlet/a/location/lakeside/rig/v001/p'), Sid('asset__state:hamlet/a/location/lakeside/rig/v002/p'), Sid('asset__state:hamlet/a/location/queens_chamber/rig/v001/p'), Sid('asset__state:hamlet/a/location/queens_chamber/rig/v002/p'), Sid('asset__state:hamlet/a/location/ramparts/rig/v001/p'), Sid('asset__state:hamlet/a/location/ramparts/rig/v002/p')]

        Combining parent and constant search

            >>> list( finder_states.find('hamlet/a/location/ramparts/rig/*/*') )
            [Sid('asset__state:hamlet/a/location/ramparts/rig/v001/w'), Sid('asset__state:hamlet/a/location/ramparts/rig/v001/p'), Sid('asset__state:hamlet/a/location/ramparts/rig/v002/w'), Sid('asset__state:hamlet/a/location/ramparts/rig/v002/p')]

        Finds itself if needed.

            >>> list( finder_states.find('hamlet/a/location/ramparts/rig/v001/p') )
            [Sid('asset__state:hamlet/a/location/ramparts/rig/v001/p')]

    This is still beta code.
    """

    def __init__(self, key: str, values: List[str], parent_source: Finder | None = None):  # noqa
        """


        Args:
            key: the keytype for which this Finder operates.
            values: the list of values that this Finder always finds, for the given key.
            parent_source: delegates to this finder for searches above the keytype (parents of the keytype)
        """
        self.key = key
        self.values = values
        self.parent_source = parent_source

    def _append_value(
        self, root: Sid, done: set, as_sid: bool = False
    ) -> Iterator[Sid] | Iterator[str]:
        """
        This method fills the keytype with the values, by appending them to the parent root.

        Args:
            root:
            done:
            as_sid:

        Returns:
        """

        for value in self.values:

            result = root.get_with(key=self.key, value=value)
            if not result:
                debug(f'Generated Sid "{result}" is not valid, skipped (used {self.key}:{value})')
                continue

            if result not in done:
                # done.add(result)
                if as_sid:
                    yield result
                else:
                    yield str(result)

    def star_search(
        self, search_sids: List[Sid], as_sid: bool = False, do_sort: bool = False
    ) -> Iterator[Sid] | Iterator[str]:

        done = set()

        for search_sid in search_sids:

            root = Sid(search_sid).get_as(self.key)

            if not root:
                continue

            # nothing to search, we yield
            if "*" not in str(root):

                if root not in done:
                    # done.add(root)  # TODO: useful ?
                    if as_sid:
                        yield root
                    else:
                        yield str(root)

            # the parent needs to be searched
            elif "*" in str(root.parent) and root != root.parent:

                if not self.parent_source:
                    raise SpilException(
                        f"The parent of {root} is a search, but no parent_source was given."
                    )

                for found_root in self.parent_source.find(root.parent):

                    # the constant "key" is not searched
                    # we can simply yield the parent and the key's value
                    if root.get(self.key) != "*":
                        result = found_root / root.get(self.key)
                        if result not in done:
                            # done.add(result)
                            if as_sid:
                                yield result
                            else:
                                yield str(result)

                    # the "key" is a search, so we append the constant values
                    else:
                        generator = self._append_value(found_root, done, as_sid=as_sid)
                        yield from generator

            # no parent search, we just need to append the constant values
            else:
                generator = self._append_value(root, done, as_sid=as_sid)
                yield from generator

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- Key: "{self.key}" -- values: {self.values} -- parent source: \n\t"{self.parent_source}"]'


if __name__ == "__main__":

    from spil.conf import projects, asset_types
    from spil.util.log import setLevel, WARN

    setLevel(WARN)

    cs1 = FindInConstants("project", projects)
    cs2 = FindInConstants("type", ["a", "s"], parent_source=cs1)
    cs3 = FindInConstants("assettype", asset_types, parent_source=cs2)
    cs4 = FindInConstants("sequence", ["sq088"], parent_source=cs2)
    print(cs2)

    for i in cs1.find("*"):
        print(f"sid: {i} / {type(i)} / path: {i.path()}")
    print("-" * 20)

    for i in cs2.find("hamlet/*"):
        print(f"sid: {i} / {type(i)} / path: {i.path()}")
    print("-" * 20)

    for i in cs3.find("hamlet/*/*"):
        print(f"sid: {i} / {type(i)} / path: {i.path()}")
    print("-" * 20)

    for i in cs4.find("hamlet/*/*"):
        print(f"sid: {i} / {type(i)} / path: {i.path()}")
    print("-" * 20)



# -*- coding: utf-8 -*-
"""

This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2022 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
from spil.sid.core.utils import extrapolate


def execute(sids):
    """
    Extrapolates the searches.

    :param sids:
    :return:
    """
    result = []
    for sid in sids:
        result.extend(extrapolate([sid], as_sid=True))

    return result




"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Tuple, List, Optional, Mapping

import string
from collections import OrderedDict

import spil.vendor  # SMELL
import lucidity  # type: ignore
from lucidity import Template  # type: ignore

from spil.util.caching import lru_kw_cache as cache
from spil.util.log import debug, info
from spil.util.exception import SpilException, raiser

# sid conf
from spil.conf import key_types, sidtype_keytype_sep  # type: ignore
from spil.conf import sip, sid_templates  # type: ignore

"""
Sid resolver
Transforms the sid string into a valid sid dict, and reverse.
"""


@cache
def sid_to_dict(sid: str, _type: Optional[str] = None) -> Tuple[str, dict] | Tuple[None, None]:
    """
    Parses a given "sid" string using the existing sid_config templates.
    If "_type" is given, only the template named after the given "_type" is parsed.

    Returns a tuple with the type and the resolved data dict.
    If the parsing failed (no template matching) returns a None, None tuple.

    This function parses all templates in one go, so data from the first matching template is returned.
    This is the normal usage (as opposed to sid_to_dicts)

    Args:
        sid: a Sid string
        _type: a Sid type name, which is the key to the template in spil_sid_conf.

    Returns: a tuple with the type and the resolved data dict.

    """
    if _type:
        template = Template(
            _type,
            sid_templates.get(_type) or raiser(f"resolver received an invalid type: {_type}"),
            anchor=lucidity.Template.ANCHOR_BOTH,
            default_placeholder_expression="[^/]*",
            duplicate_placeholder_mode=lucidity.Template.STRICT,
        )

        templates = [template]

    else:
        templates = []
        for name, pattern in sid_templates.items():
            template = Template(
                name,
                pattern,
                anchor=lucidity.Template.ANCHOR_BOTH,
                default_placeholder_expression="[^/]*",
                duplicate_placeholder_mode=lucidity.Template.STRICT,
            )

            templates.append(template)

    try:
        data, template = lucidity.parse(sid, templates)
        # print 'found', data, template
    except lucidity.ParseError as e:
        debug(f'Lucidity did not find a matching pattern. Type: {_type} (Message: "{e}")')
        return None, None

    if not data:
        return None, None

    # Sorting the result data into an OrderedDict()
    sid_type = template.name.split(sidtype_keytype_sep)[0]
    keys = key_types.get(sid_type)  # using template to get sorted keys
    keys = filter(lambda x: x in template.keys(), keys)  # template.keys() is a set

    data = data.copy()
    ordered = OrderedDict()
    for key in keys:
        ordered[key] = data.get(key)

    return template.name, ordered


@cache
def sid_to_dicts(sid: str) -> Mapping[str, dict]:
    """
    Parses a given "sid" using the existing spil_sid_config templates.
    Returns a dict with the types and the resolved data.

    This function parses all templates separately, so data from all matching templates are returned.
    This usage is not standard (as opposed to sid_to_dict).
    It makes sense only for broad read sids (including /**) where we want to catch-all types.

    Args:
        sid: a Sid string

    Returns: a dictionary containing types as keys and resolved data as values.

    """
    results = {}

    # instantiating lucidity Templates
    for name, pattern in sid_templates.items():
        template = Template(
            name,
            pattern,
            anchor=lucidity.Template.ANCHOR_BOTH,
            default_placeholder_expression="[^/]*",  # allows empty keys
            duplicate_placeholder_mode=lucidity.Template.STRICT,
        )

        # try template parse
        try:
            data, template = lucidity.parse(sid, [template])
        except lucidity.ParseError as e:
            # ParseErrors are normal, we force the parsing of all the templates,
            # and not just the first that matches, as usually
            continue

        if not data:
            continue

        # Sorting the result data into an OrderedDict():
        # getting the type
        sid_type = template.name.split(sidtype_keytype_sep)[0]
        # using template to get sorted keys for this basetype
        keys = key_types.get(sid_type)
        # filter out unused keys
        keys = filter(lambda x: x in template.keys(), keys)
        # building an Ordered dict with keys in right order
        data = data.copy()
        ordered = OrderedDict()
        for key in keys:
            ordered[key] = data.get(key)
        # appending the oredered dict to the results list
        results[template.name] = ordered

    return results


def dict_to_sid(data: dict, _type: Optional[str] = None) -> str:
    """
    Formats the given "data" dictionary using the given template "_type".
    If "_type" is not given it is detected using "dict_to_type".

    Returns the sid string.

    Args:
        data: a data dictionary with Sid fields
        _type: a Sid type name, which is the key to the template in spil_sid_conf.

    Returns: a Sid string

    """
    if not data:
        raise SpilException("[dict_to_sid] Data is empty")

    data = data.copy()

    if not _type:
        _type = dict_to_type(data)  # type: ignore

    pattern = sid_templates.get(_type)

    if not pattern:
        raise SpilException(f'Unable to find pattern for type: "{_type}" \nData: "{data}"')

    template = lucidity.Template(_type, pattern)

    if not template:
        raise SpilException(f"Unexpected: No template for type: {_type}")
    try:
        sid = template.format(data).rstrip(sip)
    except lucidity.error.FormatError as e:
        debug(f'Lucidity could not format the Sid. Data: {data} / type: {_type} (Message: "{e}")')
        return ""

    return sid


# SMELL - this code is obscure and should be replaced / not be used
def dict_to_type(data: dict, all: bool = False) -> str | List[str]:
    """
    Retrieves the sid types for the given dict "data".
    "data" can be unsorted.

    If "all" is False (default), the first matching type is returned.
    If "all" is set to True, of list of corresponding types is returned,
    although usually a single type should match.

    The types are found by:
    - comparing the keys of the dict with the keys of the existing templates
    - in case of matching keys, applying the template and resolving it back to a dict

    Multiple matching types is a sign for a configuration problem.
    It is logged using debug('Sid multitypes for  =>')

    Args:
        data: a data dictionary with Sid fields
        all: if set to True, all matching types are returned, else the first matching type is returned (default)

    Returns: the type or types resolved from the given data dictionary.
    """

    found = []
    keys = data.keys()

    # SMELL: this whole code is obscure...
    for _type, template in sid_templates.copy().items():

        template_keys = [t[1] for t in string.Formatter().parse(template) if t[1] is not None]

        if len(keys) == len(template_keys):
            if set(keys) == set(template_keys):
                ltemplate = lucidity.Template(_type, template)
                test = ltemplate.format(data)
                if test:
                    # FIXME / #SMELL : this code is plain nonsense...
                    debug("Checking matching types ... (fails are normal)")
                    a, b = sid_to_dict(test, _type)
                    if a:
                        # print _type
                        found.append(_type)

    if not found:
        info("No type found for {}".format(data))
        return ""

    if len(found) > 1:
        debug("Sid multitypes for  => {} // {}".format(data, found))

    if all:
        return found
    else:
        return found[0]


if __name__ == "__main__":

    from spil.util.log import setLevel, INFO, info, warning

    info("Tests start")

    setLevel(INFO)
    # setLevel(DEBUG)  # In case of problems, use DEBUG mode

    tests = ["hamlet/s/*"]

    for test in tests:

        info("*" * 15)
        info("Testing : {}".format(test))

        __, _dict = sid_to_dict(test)
        info("sid {} ---> \n{}".format(test, _dict))

        a_type = dict_to_type(_dict)
        info("type : {}".format(a_type))

        info("------ keys : {}".format(_dict.keys()))

        retour = dict_to_sid(_dict)
        info("retour: " + retour)

        assert test == retour

        info("*" * 15)
        print("  ")

    info("*" * 30)
    info("*" * 30)



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from pathlib import Path


def write_sids_to_file(sids, sid_file):
    """
    This function writes given Sids to the given Sid file
    (EOL separated file = one Sid string per line).

    To obtain Example Sids by parsing the existing files, use parse_sid_files.py.
    """
    sid_file = Path(sid_file)

    if sid_file.exists():
        raise Exception('The test file "{}" already exists. Skipped'.format(sid_file))

    with open(str(sid_file), 'w') as f:
        for sid in sids:
            f.write(str(sid) + '\n')

    print('Written {} Sids to {}'.format(len(sids), sid_file))


if __name__ == '__main__':

    print('Generating Example Sids.')

    sids = ['hamlet/s/sq010/sh0010/layout/v001/w/hip',
            'hamlet/s/sq010/sh0010/anim/v001/w/ma',
            'hamlet/s/sq010/sh0010/render/v001/p/mov']

    # sid_file = ""  # define path
    # write_sids_to_file(sids, sid_file)

    print('Done.')



# type: ignore
"""
This script generates example Sids, to test the configuration and the SPIL features towards the data.

It is recommended to generate a test script like this when configuring spil.
It helps identify the types and eliminate ambiguities.

The script can also launch a "test_config_coverage",
showing if all the types covered by the test sids,
and which types are not.

For testing, these Sids can be copied into data/testing/hamlet.sids.txt
This file is then used as test input.

Note:
The current settings are used for testing, changing them impacts the test results.
"""
import random
from math import ceil
from time import sleep

from spil_sid_conf import projects, asset_types, asset_tasks
from spil_sid_conf import extensions_scene, extensions_movie, shot_tasks, extensions_cache
from logging import debug

from spil.tests.config_checks.check_03_config_coverage import test_config_coverage

do_intermediates = True  # If this is False, only leave paths will be generated. See: LS extrapolate.
repeat_times = 1  # 3
num_sequences = 3  # 12
num_shots = 4  # 24
num_versions = 2  # 12

sid_types = ['a', 's']

asset_names = {
                'char': ['hamlet', 'claudius', 'ghost', 'ophelia', 'gertrude', 'horatio', 'polonius'],
                'location': ['elsinore', 'ramparts', 'queens_chamber', 'garden', 'lakeside'],
                'prop': ['sword', 'throne', 'dagger', 'skull', 'crown', 'tapestry'],
                'fx': ['mist', 'blood', 'thunder', 'rain', 'water'],
}

all_assets = sum(asset_names.values(), [])

task_extensions = {
    'art': ['psd', 'mov', 'mp4'],
    'model': ['ma', 'hip', 'blend', 'mp4'],
    'surface': ['ma', 'hip', 'blend', 'mp4'],
    'rig': ['ma', 'hip', 'blend', 'mp4'],
    'board': ['psd', 'mov'],
    'layout': ['ma', 'mb', 'hip', 'blend', 'mp4'],
    'anim': ['mb', 'hip', 'abc', 'mp4'],
    'fx': ['hip', 'abc', 'vdb', 'mp4', 'mov'],
    'render': ['ma', 'mb', 'hip', 'blend', 'mp4', 'mov', 'avi'],
    'comp': ['nk', 'mp4', 'mov'],
}

states = ['w', 'p']

extensions = extensions_scene + extensions_movie + extensions_cache

sequences = ['sq{}'.format(str(x*10). zfill(3)) for x in range(1, num_sequences+1)]
shots = ['sh{}'.format(str(x*10). zfill(4)) for x in range(1, num_shots+1)]
versions = ['v{}'.format(str(x). zfill(3)) for x in range(1, num_versions+1)]

random.seed(26)


def append(sids, data):
    sid = '/'.join(data)
    sids.append(sid)


sids = []
for i in range(repeat_times):
    for project in projects:
        debug('Generating project: {}'.format(project))
        if do_intermediates: append(sids, [project])

        for sid_type in sid_types:
            debug('Generating type: {}'.format(sid_type))
            data = [project, sid_type]
            if do_intermediates: append(sids, data)

            if sid_type in ['a']:

                for assettype in asset_types:
                    data = [project, sid_type, assettype]
                    if do_intermediates: append(sids, data)

                    for name in asset_names.get(assettype):
                        data = [project, sid_type, assettype, name]
                        if do_intermediates: append(sids, data)

                        for task in asset_tasks:
                            data = [project, sid_type, assettype, name, task]
                            if do_intermediates: append(sids, data)

                            for version in versions:
                                data = [project, sid_type, assettype, name, task, version]
                                if do_intermediates: append(sids, data)

                                for state in states:
                                    data = [project, sid_type, assettype, name, task, version, state]
                                    if do_intermediates: append(sids, data)

                                    k = ceil(len(task_extensions.get(task)) / 2)
                                    for ext in random.sample(task_extensions.get(task), k=k):  # random.choices >py 3.6
                                        data = [project, sid_type, assettype, name, task, version, state, ext]
                                        append(sids, data)

            if sid_type in ['s']:

                for seq in sequences:
                    data = [project, sid_type, seq]
                    if do_intermediates: append(sids, data)

                    for shot in shots:
                        data = [project, sid_type, seq, shot]
                        if do_intermediates: append(sids, data)

                        for task in shot_tasks:
                            data = [project, sid_type, seq, shot, task]
                            if do_intermediates: append(sids, data)

                            for version in versions:
                                data = [project, sid_type, seq, shot, task, version]
                                if do_intermediates: append(sids, data)

                                for state in states:
                                    data = [project, sid_type, seq, shot, task, version, state]
                                    if do_intermediates: append(sids, data)

                                    k = ceil(len(task_extensions.get(task)) / 2)
                                    for ext in random.sample(task_extensions.get(task), k=k):
                                        if ext in extensions_cache and state == 'p':
                                            node = random.choice(all_assets)
                                            data = [project, sid_type, seq, shot, task, version, state, node, ext]
                                        else:
                                            data = [project, sid_type, seq, shot, task, version, state, ext]
                                        append(sids, data)

sids = sorted(list(set(sids)))


if __name__ == '__main__':

    import sys
    # from tests import Timer

    from spil.util.log import DEBUG, get_logger

    log = get_logger("spil_tests")
    log.setLevel(DEBUG)

    from spil import Sid

    print('Start')
    #pprint(sids)

    #global_timer = Timer(name="global", logger=log.debug)
    #global_timer.start()

    print(len(sids))

    if True:  # set False to skip
        test_config_coverage(sids)

    y = input("print out detail ?")

    if y != 'y':
        sys.exit()

    for i, s in enumerate(sids):
        sid = Sid(s)
        if not sid:
            print(f"---------------------------------> {sid}")
            sleep(1)
        if i % 20 == 0:
            print(f"{i} -- {repr(sid)}",)

    # global_timer.stop()
    print('Done')

    print(len(sids))




"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.util.log import DEBUG, ERROR, get_logger
from spil import Sid, SpilException
from spil import FindInPaths  # , Data
from pprint import pformat

log = get_logger("spil_tests")
log.setLevel(DEBUG)


def test_full_sid(s, reraise=True, replace=None, from_search=None):
    """
    Test protocol for the full Sid.

    Only major or unexpected problems trigger exceptions.
    Most configuration problems trigger log warnings.

    If reraise is True, exceptions are reraised, otherwise there are logged.
    replace can be a tuple that is used for string replacement on the tested Sid.

    from_search allows to pass a search string, to test the match.
    """
    s = str(s)
    if replace:
        s = s.replace(*replace)

    log.info('Testing: "{}"'.format(s))
    sid = Sid(s)
    log.info('Instanced: "{}"'.format(sid.uri))

    if not s.count("?"):  # Assert works only without Query part
        assert str(sid) == s
    assert sid == eval(repr(sid))

    try:
        if not sid.type:
            log.error('Sid "{}" not typed, skipping'.format(sid))
            return

        test_search(sid)

        if from_search:
            match = sid.match(from_search)
            if not match:
                log.warning('No match "{}" <-> "{}". This is not normal'.format(sid, from_search))

        key = sid.keytype
        parent_key = sid.parent.keytype
        if not sid.type == "project":
            if sid.parent and sid.get_as(parent_key):
                if not sid.parent == sid.get_as(parent_key):
                    log.warning('Sid "{}" parent problem'.format(sid))
            else:
                log.warning('Sid "{}" has not parent ?'.format(sid))

        assert sid == Sid(
            sid.get("project") + "?" + sid.as_query()
        ), "Sid Query assertion pb : {}".format(Sid(sid.get("project") + "?" + sid.as_query()()))

        assert sid == Sid(
            sid.get("project") + "?" + sid.as_query()
        ), "Sid Query assertion pb : {}".format(Sid(sid.get("project") + "?" + sid.as_query()()))
        log.info('Passed "Query": sid == Sid(sid.get("project") + "?" + sid.as_query() ')

        if not sid.parent:
            log.warning("Sid {} has no valid parent (got: {}).".format(sid, sid.parent))
        elif sid == sid.parent:
            log.info("Sid is a root (has no parent).")
        else:
            assert sid == sid.parent / sid.get(sid.keytype)
            log.info('Passed "parent": sid == sid.parent / sid.get(sid.keytype) ')

        path = sid.path()
        if path:
            try:
                assert sid == Sid(path=sid.path())
                log.info("Passed : sid == Sid(path=sid.path()) ")
            except AssertionError:
                msg = "Sid path is ambiguous. {}\nsid: {}\nsid.path(): {}\nSid(path=sid.path()): {}\n".format(
                    "(Sid is a search, this may be normal)." if sid.is_search() else "",
                    sid,
                    sid.path(),
                    Sid(path=sid.path()),
                )
                log.warning(msg)
            try:
                assert sid.path() == Sid(path=sid.path()).path()
                log.info("Passed : sid.path() == Sid(path=sid.path()).path() ")
            except AssertionError:
                msg = "Sid path is ambiguous. {}\nsid: {}\nsid.path(): {}\nSid(path=sid.path()): {}\nSid(path=sid.path()).path(): {}".format(
                    "(Sid is a search, this may be normal)." if sid.is_search() else "",
                    sid,
                    sid.path(),
                    Sid(path=sid.path()),
                    Sid(path=sid.path()).path(),
                )
                log.warning(msg)
        else:
            log.error('Sid "{}" has no path.'.format(sid))

        params = {
            "parent": sid.parent,
            "grand_parent": sid.parent.parent,
            "basetype": sid.basetype,
            "keytype": sid.keytype,
            "path": sid.path(),
            "exists": sid.exists(),
            "is_leaf": sid.is_leaf(),
            "len": len(sid),
            "type": sid.type,
            "uri": sid.uri,
            "string": sid.string,
            "query": sid.as_query(),
            "is_search": sid.is_search(),
            "get_last": sid.get_last(),
            "children": list(sid.children()),
            "siblings": list(sid.siblings()),
            "uncles": list(sid.parent.siblings()),
        }

        if sid.get("version"):
            params.update(
                {
                    "get_last (version)": sid.get_last("version"),
                    "get_next (version)": sid.get_next("version"),
                    "get_new (version)": sid.get_new("version"),
                }
            )

        log.info("Params: \n" + pformat(params))
        log.info("Fields: \n" + pformat(sid.fields))

    except SpilException as e:
        log.error("SpilException : {} --> {}".format(e, s))
        if reraise:
            raise e

    except Exception as e:
        log.error("Exception : {} --> {}".format(e, s))
        if reraise:
            raise e


def test_full_sids(sids, reraise=True, replace=None):
    """
    Loop over test_full_sid.
    """

    log.info("Testing if example sids match the Sid config_name")

    if not sids:
        log.warning("No sids given, nothing to test.")
        return

    for i, s in enumerate(sids):
        log.info("*" * 75)
        log.info("----------------------------------------- {}".format(i))
        test_full_sid(s, reraise=reraise, replace=replace)


def test_search(sid):
    """
    Runs a search on the given Sid.
    If the Sid exists on the system, the search result should equal the search.
    This test is called by test_sid.
    """

    sid = Sid(sid)

    found = FindInPaths().find_one(sid)
    log.info("{} ----> FindInPaths() --->  {}".format(sid, found))
    if found and not sid.is_search():
        assert sid == found

    """
    found = Data().get_one(sid)
    log.info("{} ----> Data() --->  {}".format(sid, found))
    # assert(sid == found)
    """


if __name__ == "__main__":

    from spil.util.log import setLevel, DEBUG, INFO

    setLevel(INFO)

    from scripts.example_sids import sids

    test_full_sids(sids)



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, Mapping, Any, List, Optional, Callable

from spil import Sid
from spil.sid.read.util import first
from spil.sid.read.tools import unfold_search


class Getter:
    """
    Interface for Sid Getter.

    Whereas the Finder only finds Sids, the Getter retrieves Sid(s) and its data.

    A Getter is getting from a data source.
    Eg. GetFromPaths, GetFromSG.

    The Getters work has two forms:
    Finding with data form a search sid, or Getting data from a Sid.

    1. Finding with data

      The Getter receives a Search Sid, finds the result Sid(s) and returns them together with data.
      This can be considered like a Finder including dependencies (the "Finder" itself only finds Sids).

      Implementation options

        - The Getter calls the data source directly and retrieves Sids with data in one go. Typically, it is done when the data source is a database (for example GetFromSG)

        - The Getter calls a Finder to find the Sids, and then retrieves the data for each of them. For example GetFromPaths calls FindInPaths and then reads the data from disk for the Sids.

      Methods: get(), do_get(), get_one()
      These methods yield over dicts containing the data.
      One special field named "sid" contains the Sid.

    2. Getting data

      The Getter receives a specific Sid, and retrieves its data.

      It is possible to call

        - For a given existing Sid, retrieving data with all or some attributes

        - For a given existing Sid, retrieving one attribute

      Methods: get_data(), get_attr()
      get_data() returns the data dict
      get_attr() returns the attributee

    """

    def get(
        self,
        search_sid: str | Sid,
        attributes: Optional[List[str]] = None,
        sid_encode: Callable = str,
    ) -> Iterator[Mapping[str, Any]]:
        """
        For a given search, returns an Iterator over Mappings containing the retrieved data.
        One special field named "sid" contains the Sid.
        The Sid can be encoded by providing a callable to "sid_encode". Default is str.

        By default, attributes is None, retrieved data contains all configured data for the Sid type.
        If "attributes" is given, data contains only the key:value for the given attributes.

        The Sids returned by Getter.get() should be identical to those returned by Finder.find().
        Getter retrieves data related to these Sids, whereas the Finder finds only the existing Sids themselves.

        Args:
            search_sid: the search sid (possibly untyped).
            attributes: a list of attribute names, that should be fetched.
            sid_encode: a callable to which the sid object will be passed, that should return the value added to the data dictionary under the "sid" key.
                        Example: `sid_encode=lambda x: x.uri`
                        If the callable returns None, the value is not added. As in `sid_encode=lambda x: None`

        Returns:
            Iterator over Mappings containing the retrieved data.
            One special field named "sid" contains the Sid
        """
        # shortcut if Sid is not a search
        sid = Sid(search_sid)
        if sid and not sid.is_search():
            generator = self.do_get([sid], attributes=attributes, sid_encode=sid_encode)
        else:
            search_sids = unfold_search(search_sid)
            generator = self.do_get(search_sids, attributes=attributes, sid_encode=sid_encode)
        yield from generator

    def do_get(
        self,
        search_sids: List[Sid],
        attributes: Optional[List[str]] = None,
        sid_encode: Callable = str,
    ) -> Iterator[Mapping[str, Any]]:
        """
        Has the same purpose as get(), but operates on given "search_sids",
        a list of typed search sids.

        Is called by get() after unfolding and typing the search sid.

        Args:
            search_sids: List of typed search sids.
            attributes: a list of attribute names, that should be fetched.
            sid_encode: a callable to which the sid object will be passed, that should return the value added to the data dictionary under the "sid" key.
                        Example: `sid_encode=lambda x: x.uri`
                        If the callable returns None, the value is not added. As in `sid_encode=lambda x: None`

        Returns:
            Iterator over Mappings containing the retrieved data.
            One special field named "sid" contains the Sid.
        """
        raise NotImplementedError("Method do_get() needs to be implemented.")

    def get_one(
        self,
        search_sid: str | Sid,
        attributes: Optional[List[str]] = None,
        sid_encode: Callable = str,
    ) -> Mapping[str, Any]:
        """
        Calls get() with the given parameters, and returns the first result.

        Args:
            See get()

        Returns:
            Mapping containing the retrieved data.
            One special field named "sid" contains the Sid
        """
        found = first(self.get(search_sid, attributes=attributes, sid_encode=sid_encode))
        return found or {}

    def get_data(
        self, sid: str | Sid, attributes: Optional[List[str]] = None, sid_encode: Callable = str
    ) -> Mapping[str, Any]:
        """
        Returns data associated to the given Sid, as a key:value dictionary.
        One special field named "sid" contains the Sid.

        Returns an empty dictionary if no data was found.

        Args:
            sid: a typed sid
            attributes: a list of attribute names, that should be fetched.
            sid_encode: a callable to which the sid object will be passed, that should return the value added to the data dictionary under the "sid" key.
                        Example: `sid_encode=lambda x: x.uri`
                        If the callable returns None, the value is not added. As in `sid_encode=lambda x: None`

        Returns:
            Mapping containing the retrieved data.
            One special field named "sid" contains the Sid
        """
        return self.get_one(search_sid=sid, attributes=attributes, sid_encode=sid_encode)

    def get_attr(self, sid: str | Sid, attribute: str) -> Any | None:
        """
        Returns an attribute, a named piece of data, for the given Sid.

        Args:
            sid: typed Sid
            attribute: attribute name

        Returns:
            The value for the given Sid and attribute name.
        """
        return self.get_data(sid).get(attribute)

    def __str__(self):
        return f"[spil.{self.__class__.__name__}]"


if __name__ == "__main__":
    print(Getter())



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Tuple, Optional

import os
from pathlib import Path
from collections import OrderedDict  # TODO: replace by dict

import spil.vendor  # SMELL
import lucidity  # type: ignore
from lucidity import Template  # type: ignore

from spil.util.caching import lru_kw_cache as cache
from spil.util.log import debug
from spil.util.exception import SpilException

from spil.util import utils
from spil.conf import key_types, sidtype_keytype_sep
from spil.sid.core.sid_resolver import dict_to_type, sid_to_dict
from spil.sid.pathops.pathconfig import get_path_config, PathConfig

"""
File system resolver
Path <-> dict translation
"""


def get_resolvers(config: PathConfig) -> dict:
    """
    Technical necessity for Lucidity template reference mechanism.

    Args:
        config:

    Returns:

    """
    return {
        config.path_templates_reference: Template(  # type: ignore
            config.path_templates_reference, config.path_templates.get(config.path_templates_reference)  # type: ignore
        )
    }


@cache
def path_to_dict(
    path: str | os.PathLike[str], _type: Optional[str] = None, config: Optional[str] = None
) -> Tuple[str, dict] | Tuple[None, None]:
    """
    Resolves the given path into the matching data dictionary.
    Uses the _type, if given, else looks up all templates and uses the first matching one.

    Uses the Lucidity template mechanism,
    then applies configured mappings.

    Returns a tuple with the type and the parsed data in an OrderedDict.
    If the parsing failed (no template matching) returns a None, None tuple.

    """
    pc = get_path_config(config)
    resolvers = get_resolvers(pc)

    path = str(path)
    path = path.replace(os.sep, "/")

    if _type:
        template = Template(
            _type,
            pc.path_templates.get(_type),
            anchor=lucidity.Template.ANCHOR_BOTH,
            default_placeholder_expression="[^/]*",
            duplicate_placeholder_mode=lucidity.Template.STRICT,
        )
        template.template_resolver = resolvers
        templates = [template]

    else:
        templates = []
        for name, pattern in pc.path_templates.items():
            template = Template(
                name,
                pattern,
                anchor=lucidity.Template.ANCHOR_BOTH,
                default_placeholder_expression="[^/]*",
                duplicate_placeholder_mode=lucidity.Template.STRICT,
            )
            template.template_resolver = resolvers
            templates.append(template)

    try:
        data, template = lucidity.parse(path, templates)
        # print 'found', data, template
    except lucidity.ParseError as e:
        debug(f'Lucidity did not find a matching pattern. Type given: {_type} (Message: "{e}")')
        return None, None

    if not data:
        return None, None

    # path mapping
    for key, value in data.items():
        # debug('{}, {}, {}'.format(key, value, template.name))
        if pc.path_mapping.get(key):
            value = pc.path_mapping.get(key).get(value, value)
            data[key] = value

            mapping = pc.path_mapping.get((key, template.name))  # type specific mapping
            if mapping:
                data[key] = mapping.get(value, value)

    # mapping from extra keys
    for new_key, sid_mapping in pc.extrakeys_to_sidkeys.items():
        if new_key in data:
            for key, _dict in sid_mapping.items():
                map_result = _dict.get(data.get(new_key))
                if map_result:
                    data[key] = map_result

    # Sorting the result data into an OrderedDict()
    sid_type = template.name.split(sidtype_keytype_sep)[0]
    keys = key_types.get(sid_type)  # using template to get sorted keys
    keys = filter(lambda x: x in data.keys(), keys)  # template.keys() is a set #

    data = data.copy()
    ordered = OrderedDict()
    for key in keys:
        ordered[key] = data.get(key)

    return template.name, ordered


def dict_to_path(data: dict, _type: Optional[str] = None, config: Optional[str] = None) -> Path:
    """
    Resolves the given data dictionary into a path.
    Uses the _type, if given, else calls dict_to_type to find matching type.

    Uses the Lucidity template mechanism.
    Applies configured mappings and defaults before template formatting.

    Returns path string.
    """
    if not data:
        raise SpilException("[dict_to_path] Data is empty")

    pc = get_path_config(config)
    resolvers = get_resolvers(pc)

    data = data.copy()

    debug(f"Data: {data}")

    # setting defaults on empty values
    for key in data.keys():
        if not data.get(key) and pc.path_defaults.get(key):
            data[key] = pc.path_defaults.get(key)

    if not _type:
        _type = dict_to_type(data)

    # reverse path mapping
    for key, value in data.items():
        if value and pc.path_mapping.get(key):
            mapping = pc.path_mapping.get(key)  # global mapping
            data[key] = utils.get_key(mapping, value, value)

            mapping = pc.path_mapping.get((key, _type))  # type specific mapping
            if mapping:
                data[key] = utils.get_key(mapping, value, value)

    debug(f"sidtype: {_type}")

    pattern = pc.path_templates.get(_type)

    debug(f"pattern: {pattern}")

    if not pattern:
        raise SpilException(f'Unable to find pattern for type: "{_type}" \nData: "{data}"')

    template = Template(_type, pattern)
    template.template_resolver = resolvers

    debug(f"template: {template}")

    if not template:
        raise SpilException(f"Unexpected: No template for type: {_type}")

    # adding template specific defaults
    for key in template.keys():
        if key not in data.keys() and pc.path_defaults.get(key):
            data[key] = pc.path_defaults.get(key)

    # adding extra keys
    for key in pc.sidkeys_to_extrakeys.keys():
        if key in data.keys():
            for new_key, mapping in pc.sidkeys_to_extrakeys.get(key, {}).items():
                data[new_key] = mapping.get(data.get(key))

    debug(f"data after path_defaults: {data}")
    path = template.format(data)

    debug(f"found: {path}")

    return Path(path)


if __name__ == "__main__":

    from spil.util.log import setLevel, INFO, info, warning

    info("Tests start")

    setLevel(INFO)
    # setLevel(DEBUG)  # In case of problems, use DEBUG mode

    tests = ["hamlet/s/*"]

    for test in tests:

        info("*" * 15)
        info("Testing : {}".format(test))

        _type, _dict = sid_to_dict(test)
        info("Dict ({}) ---> \n{}".format(_type, _dict))

        try:
            path = dict_to_path(_dict, _type=_type, config="server")
            info(f"Path: {path}")

            __, retour = path_to_dict(path, config="server")
            info("Retour ({}) ---> \n{}".format(_type, retour))

            assert _dict == retour

        except SpilException as se:
            warning(se)

        info("*" * 15)
        print("  ")

    info("*" * 30)
    info("*" * 30)



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from typing import List

from spil.conf import sip, ors
from spil.sid.core import query_helper


def execute(sids: List[str]) -> List[str]:
    """
    Runs or_op (the "or operator") on a list of Sids.

    Args:
        sids: a list of Sid strings to edit

    Returns: the list of Sid strings, edited
    """
    result = []
    for sid in sids:
        result.extend(or_op(sid))

    return result


def or_op(sid: str) -> List[str]:
    """
    or_op (the "or operator") transforms a string containing the "or" sign
    into a list of strings, each one individually representing the options without the or.

    Note: the ors sign can be configured, it is called "ors" in the config.
    Typically it is a comma ",".

    Example:

        >>> or_op('bla/s/bla/A,B/**/one,two?test=X,Y,Z')
        ['bla/s/bla/A/**/one?test=X', 'bla/s/bla/A/**/one?test=Y', 'bla/s/bla/A/**/one?test=Z', 'bla/s/bla/B/**/one?test=X', 'bla/s/bla/B/**/one?test=Y', 'bla/s/bla/B/**/one?test=Z', 'bla/s/bla/A/**/two?test=X', 'bla/s/bla/A/**/two?test=Y', 'bla/s/bla/A/**/two?test=Z', 'bla/s/bla/B/**/two?test=X', 'bla/s/bla/B/**/two?test=Y', 'bla/s/bla/B/**/two?test=Z']

    Args:
        sid: a sid string

    Returns: a list of Sid strings
    """
    sid = str(sid)
    if not sid.count(ors):  # no "or" operators in sid.
        return [sid]

    if sid.count("?"):  # sid contains Query ending. We put it aside, and later append it back
        sid, query = sid.split("?", 1)
    else:
        query = ""

    sids = or_on_path(sid)

    result = []
    if query:
        uris = or_on_query(query)
        for s in sids:
            for u in uris:
                result.append("{}?{}".format(s, u))
    else:
        result = sids

    return result


def or_on_path(sid):
    """
    Applies the or_op on the path part of the Sid.

    Example:

        >>> or_on_path('bla/s/bla/A,B,C/**/one,two,three')
        ['bla/s/bla/A/**/one', 'bla/s/bla/B/**/one', 'bla/s/bla/C/**/one', 'bla/s/bla/A/**/two', 'bla/s/bla/B/**/two', 'bla/s/bla/C/**/two', 'bla/s/bla/A/**/three', 'bla/s/bla/B/**/three', 'bla/s/bla/C/**/three']

    Args:
        sid: sid string

    Returns: list of sid string
    """

    _start = "--start--"

    parts = sid.split(sip)

    found = [_start]
    for part in parts:
        current = found.copy()
        if ors in part:
            for alt in part.split(ors):
                alt = alt.strip()
                for sid in current.copy():
                    new = sid + sip + alt
                    # print 'replace', sid, ' --> ', new, ' -- ', sid in found, '?'
                    #
                    if sid in found:
                        found[found.index(sid)] = new  # replace (of the first element)
                    else:
                        found.append(new)  # replace (of the first element)

            # found.remove()
        else:
            for sid in found.copy():
                new = sid + sip + part
                # print new
                found[found.index(sid)] = new  # replace (of the first element)

    result = []
    for sid in found:
        if not sid in result:
            result.append(sid.replace(_start + sip, ""))

    # no type check needed
    return result


def or_on_query(query):
    """
    Applies the or operator to values of the query, creating unique uris without the operator.

    Example:

        >>> or_on_query('titi=tata,blip&roger=vadim,bom,tom, tata')
        ['titi=tata&roger=vadim', 'titi=blip&roger=vadim', 'titi=tata&roger=bom', 'titi=blip&roger=bom', 'titi=tata&roger=tom', 'titi=blip&roger=tom', 'titi=tata&roger=tata', 'titi=blip&roger=tata']

    Args:
        query:

    Returns:

    """
    query_dict = query_helper.to_dict(query)
    result = [query_dict.copy()]
    for key, value in query_dict.items():
        if value.count(ors):
            new_result = []
            for i in value.split(ors):
                for d in result.copy():
                    new_dict = d.copy()
                    new_dict[key] = i
                    new_result.append(new_dict)
            result = new_result
    #     print(result)

    return [query_helper.to_string(d) for d in result]


if __name__ == "__main__":

    import doctest

    doctest.testmod()



"""
This test runs basic, "core" tests on Sid objects.

Uses the example sids.
"""
import random

from hamlet_scripts.example_sids import sids  # type: ignore
from spil.tests.utils.sid_core_tester import check_typed_sids
from spil.util.log import setLevel, INFO

def test_typed_sids():

    random.seed(26)
    random.shuffle(sids)
    tests = sids[:100]
    check_typed_sids(tests)


if __name__ == "__main__":

    setLevel(INFO)
    test_typed_sids()



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.util.log import debug, info
from spil import Sid, SpilException
from spil.sid.pathops.write_paths import WriteToPaths

"""
Creates all given Sids on the filesystem, as defined in data_conf (template or touch)

Important: uses the actual project paths, as in config_name. Be careful !!

"""


def generate_mock_fs(sids, config='fs_conf', force=False):

    root = Sid(sids[0]).get_as('project').path(config)
    print(root)
    if root.exists() and list(root.iterdir()) and not force:
        raise SpilException(f"Root folder {root} exists and is not empty. "
                            f"This script is about to fill it with mock data. "
                            f"For safety, the script is aborted. "
                            f"Use force=True to force. ")

    writer = WriteToPaths(config)

    for sid in sids:
        try:
            writer.create(sid)
            info(f"Created {sid}")
        except SpilException as e:
            debug(e)


if __name__ == '__main__':

    from spil.util.log import setLevel, INFO

    print('')
    print('Tests start')
    print('')

    # setLevel(WARN)
    setLevel(INFO)
    # setLevel(DEBUG)  # In case of problems, use DEBUG mode

    sids = ['hamlet/s/sq010/sh0010/layout/v001/w/hip',
            'hamlet/s/sq010/sh0010/anim/v001/w/ma',
            'hamlet/s/sq010/sh0010/render/v001/p/mov']

    generate_mock_fs(sids, config='server')

